#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu May  4 21:22:42 2017

A python implementation of P&V2002 algorithm.

Original paper http://adsabs.harvard.edu/abs/2002A%26A...385.1095P
Adamow's paper http://adsabs.harvard.edu/abs/2012MSAIS..22...48A

Adapted from Monika Adamow.

"""
# %%
import numpy as np
from scipy.linalg import solve_banded


def pyREDUCE(order, ycen, osample=10, lamb_sp=0.0, lambda_sf=0.1, use_mask=0,
             noise=5.85, uncert=False, im_output=False, normflat=False,
             slitfunc=False, wing=0.2):
    # order:		order as spatial pixel by dispersion pixel array
    # ycen:		coordinates of the centre along the order in the dispersion direction
    # osample:	oversampling rate
    # lamb_sp:	spectral step
    # lamb_sf:	slitfunction step
    # use_mask:	mask - same shape as order, 1 where pixel is good, else 0. If not supplied one will be constructed (crudely)
    # noise: 	noise value with default 5.85
    # uncert:	array of uncertainties if passing them into this method
    # im_output:	do we wish to output a reconstruction at the end?
    # normflat:	are we passing in a normalised flat field for this?
    # slitfunc:	do we wish to output the slitfunctions at the end?

    # Obtain dimensions of the input order to be reduced
    nrow, ncol = order.shape
    n = (nrow + 1) * osample + 1

    # Mask creation
    try:
        if (use_mask == 0):
            mask = np.ones((nrow, ncol))
            mask[np.where(order > 60000)] = 0.0
    except:
        mask = np.copy(use_mask)

    y = np.arange(n) / float(osample) - 1.
    bklind = np.arange(osample + 1) + n * osample
    oind = np.arange(osample + 1) * (osample + 2)
    olind = oind[0:osample + 1]
    for m in range(osample + 1, 2 * osample + 1):
        mm = m - osample
        bklind = np.append(bklind, np.arange(osample + 1 - mm) + n * m)
        olind = np.append(olind, oind[0:osample - mm + 1] + mm)

        # Now create spectrum
    sp = np.repeat(0, ncol)
    sf = np.sum(order * mask, axis=1)

    # Construction of the first slit function and spectrum
    # Counting median widths and
    # Estimating the first part of the slit function

    sf_med = np.arange(sf.shape[0] - 4)
    for i in range(2, sf.shape[0] - 2):
        sf_med[i - 2] = np.median(sf[i - 2:i + 3])
    sf[2:sf.shape[0] - 2] = sf_med
    sf = sf / np.sum(sf)

    # We estimate the first bit of the spectrum

    sp = np.sum((order * mask) * (np.outer(sf, np.repeat(1, ncol))), axis=0)
    sp_med = np.arange(sp.shape[0] - 4)
    for i in range(2, sp.shape[0] - 2):
        sp_med[i - 2] = np.median(sp[i - 2:i + 3])
    sp[2:sp.shape[0] - 2] = sp_med
    sp = sp / np.sum(sp) * np.sum(order * mask)
    dev = np.sqrt(
        np.sum(mask * (order - np.outer(sf, sp)) ** 2) / np.sum(mask))
    mask[np.where(abs(order - np.outer(sf, sp)) > 3. * dev)] = 0.0

    # Calculate weights
    weight = 1. / np.float64(osample)

    ########################################################################
    for iter in range(1, 30):
        # build a matrix with Omega as the diagonal
        Akl = np.zeros((2 * osample + 1, n))
        Bl = np.zeros((1, n))
        omega = np.repeat(weight, osample + 1)
        for i in range(0, ncol):
            # Creating arrays by weight
            omega = np.repeat(weight, osample + 1)
            yy = y + ycen[i]
            ind = np.where((yy >= 0.0) & (yy < 1.))[0]
#            figure(); plot(yy)
#            print(yy, ind)
            i1 = ind[0]
            i2 = ind[-1]
            omega[0] = yy[i1]
            omega[-1] = 1. - yy[i2]
            o = np.outer(omega, omega)
            o[osample, osample] = o[osample, osample] + o[0, 0]
            bkl = np.zeros((2 * osample + 1, n))
            omega_t = np.reshape(o, o.shape[0] * o.shape[1])
            oo = omega_t[olind]
            for l in range(0, nrow):
                bkl_temp = np.reshape(bkl, bkl.shape[0] * bkl.shape[1])
                t = l * osample + bklind + i1
                bkl_temp[t] = oo * mask[l, i]
            bkl = np.reshape(bkl_temp, (2 * osample + 1, n))
            oo = o[osample, osample]
            for l in range(1, nrow):
                bkl[osample, l * osample + i1] = oo * mask[l, i]
            bkl[osample, nrow * osample + i1] = omega[osample] ** 2 * mask[
                nrow - 1, i]
            for m in range(0, osample):
                bkl[m, (osample - m):(n)] = bkl[2 * osample - m,
                                            0:(n - osample + m)]
            Akl = Akl + (sp[i] ** 2) * bkl
            o = np.zeros((1, n))
            for l in range(0, nrow):
                o[0, l * osample + i1:l * osample + i1 + osample + 1] = \
                    order[l, i] * weight * mask[l, i]
            for l in range(1, nrow):
                o[0, l * osample + i1] = \
                    order[l - 1, i] * omega[osample] * mask[l - 1, i] + \
                    order[l, i] * omega[0] * mask[l, i]
            o[0, i1] = order[0, i] * omega[0] * mask[0, i]
            o[0, nrow * osample + i1] = order[nrow - 1, i] * omega[osample] * \
                                        mask[nrow - 1, i]
            Bl = Bl + sp[i] * o

        tab = np.zeros((n, 2))
        lamda = lambda_sf * np.sum(Akl[osample, :]) / n
        lambda_tab = np.zeros((1, n))
        for elem in range(0, n):
            lambda_tab[0, elem] = lamda

        Akl[osample, 0] = Akl[osample, 0] + lambda_tab[0, 0]
        Akl[osample, n - 1] = Akl[osample, n - 1] + lambda_tab[0, n - 1]
        Akl[osample, 1:n - 1] = Akl[osample, 1:n - 1] + 2. * lambda_tab[0,
                                                             1:n - 1]
        Akl[osample + 1, 0:n - 1] = Akl[osample + 1, 0:n - 1] - lambda_tab[0,
                                                                0:n - 1]
        Akl[osample - 1, 1:n] = Akl[osample - 1, 1:n] - lambda_tab[0, 1:n]
        Bl = Bl.T

        x = solve_banded((osample, osample), Akl, Bl, overwrite_ab=True,
                         overwrite_b=True)
        ind0 = [np.where(x < 0)]
        x[ind0] = 0.0
        sf = x / np.sum(x) * osample
        r = np.repeat(0., sp.shape[0])
        sp_old = np.copy(sp)
        dev_new = 0.0
        # sf[sf < (sf.mean() * wing)] = 0  # set wing to 0 in P&V

        for i in range(0, ncol):
            omega = np.repeat(weight, osample)
            yy = y + ycen[i]
            ind1 = np.where((yy >= 0.0) & (yy < nrow))[0]
            i1 = ind1[0]
            i2 = ind1[-1]
            omega[0] = yy[i1]
            ssf = np.reshape(sf[i1:i2 + 1], (nrow, osample))
            o = np.dot(ssf, omega)

            yyy = nrow - yy[i2]
            o[0:nrow - 1] = o[0:nrow - 1] + ssf[1:nrow, 0] * yyy
            o[nrow - 1] = o[nrow - 1] + sf[i2] * yyy
            r[i] = np.dot((order[:, i] * mask[:, i]), o)
            sp[i] = np.sum(o ** 2 * mask[:, i])
            if (iter > 1):
                norm = r[i] / sp[i]
                j = np.where(
                    abs(order[:, i] - np.transpose(norm * o)) > 7. * dev)
                mask[j, i] = 0.0
                dev_new = dev_new + np.sum(
                    mask[:, i] * (order[:, i] - np.transpose(norm * o)) ** 2)

        if (iter > 1):
            dev = np.sqrt(noise ** 2 + dev_new / np.sum(mask))

        if (lamb_sp != 0.0):
            lamda = lamb_sp * np.sum(sp) / ncol
            ab = np.zeros((3, ncol))
            ab[0, 1:] = -lamda
            ab[2, :-1] = -lamda
            ab[1, 0] = lamda + 1.
            ab[1, -1] = lamda + 1.
            ab[1, 1:-1] = 2. * lamda + 1.
            sp = solve_banded((1, 1), ab, r / sp, overwrite_ab=False,
                              overwrite_b=False)

        else:
            sp = r / sp
        
        if iter < 2:
            sf_old = np.copy(sf)
        else:
            print(iter, np.sum(np.abs(sf-sf_old))/sf_old.mean())
        
            # Convergence
        if ((abs(sp - sp_old) / sp.max()).max() < 1E-8):
            if np.sum(np.abs(sp-sp_old))/sp_old.mean() < 1E-8:
                break

    jbad = np.array(0, dtype=np.int64)
    unc = np.repeat(0., ncol)
    im_out = np.zeros_like((order))
    slitfunc_out = np.zeros_like((order))

    # Reconstruction and uncertainties
    for i in range(0, ncol):
        omega = np.repeat(weight, osample)
        yy = y + ycen[i]
        ind1 = np.where((yy >= 0.0) & (yy < nrow))[0]
        i1 = ind1[0]
        i2 = ind1[-1]
        omega[0] = yy[i1]
        ssf = np.reshape(sf[i1:i2 + 1], (nrow, osample))
        o = np.dot(ssf, omega)
        yyy = nrow - yy[i2]
        o[0:nrow - 1] = o[0:nrow - 1] + ssf[1:nrow, 0] * yyy
        o[nrow - 1] = o[nrow - 1] + sf[i2] * yyy
        j = np.where(
            (abs(order[:, i] - np.transpose(sp[i] * o))).flatten() < 5 * dev)
        b = np.where(
            (abs(order[:, i] - np.transpose(sp[i] * o))).flatten() >= 5 * dev)
        nj = sp[j].shape[0]

        if (nj < nrow):
            jbad = np.append(jbad, nrow * i + b[0])
        if (nj > 2):
            ss = np.sum((order[j, i] - sp[i] * o[j]) ** 2)
            xx = np.sum((o[j] - np.mean(o[j])) ** 2) * (nj - 2)
            unc[i] = ss / xx
        else:
            unc[i] = 0.0
        im_out[:, i] = np.transpose(sp[i] * o)
        slitfunc_out[:, i] = np.transpose(o)

    # set wing to 0
    jbad = np.array(0, dtype=np.int64)
    unc = np.repeat(0., ncol)
    im_out = np.zeros_like((order))
    slitfunc_wing0 = np.zeros_like((order))
    sf_wing0 = np.copy(sf)
    sf_wing0[sf_wing0 < (sf_wing0.mean() * wing)] = 0

    # Reconstruction and uncertainties
    for i in range(0, ncol):
        omega = np.repeat(weight, osample)
        yy = y + ycen[i]
        ind1 = np.where((yy >= 0.0) & (yy < nrow))[0]
        i1 = ind1[0]
        i2 = ind1[-1]
        omega[0] = yy[i1]
        ssf = np.reshape(sf_wing0[i1:i2 + 1], (nrow, osample))
        o = np.dot(ssf, omega)
        yyy = nrow - yy[i2]
        o[0:nrow - 1] = o[0:nrow - 1] + ssf[1:nrow, 0] * yyy
        o[nrow - 1] = o[nrow - 1] + sf_wing0[i2] * yyy
        j = np.where(
            (abs(order[:, i] - np.transpose(sp[i] * o))).flatten() < 5 * dev)
        b = np.where(
            (abs(order[:, i] - np.transpose(sp[i] * o))).flatten() >= 5 * dev)
        nj = sp[j].shape[0]

        if (nj < nrow):
            jbad = np.append(jbad, nrow * i + b[0])
        if (nj > 2):
            ss = np.sum((order[j, i] - sp[i] * o[j]) ** 2)
            xx = np.sum((o[j] - np.mean(o[j])) ** 2) * (nj - 2)
            unc[i] = ss / xx
        else:
            unc[i] = 0.0
        im_out[:, i] = np.transpose(sp[i] * o)
        slitfunc_wing0[:, i] = np.transpose(o)

    sp_wing0 = np.copy(sp)
    r = np.repeat(0., sp.shape[0])
    dev_new = 0.0
    for i in range(0, ncol):
        omega = np.repeat(weight, osample)
        yy = y + ycen[i]
        ind1 = np.where((yy >= 0.0) & (yy < nrow))[0]
        i1 = ind1[0]
        i2 = ind1[-1]
        omega[0] = yy[i1]
        ssf = np.reshape(sf_wing0[i1:i2 + 1], (nrow, osample))
        o = np.dot(ssf, omega)

        yyy = nrow - yy[i2]
        o[0:nrow - 1] = o[0:nrow - 1] + ssf[1:nrow, 0] * yyy
        o[nrow - 1] = o[nrow - 1] + sf_wing0[i2] * yyy
        r[i] = np.dot((order[:, i] * mask[:, i]), o)
        sp_wing0[i] = np.sum(o ** 2 * mask[:, i])
        if (iter > 1):
            norm = r[i] / sp[i]
            j = np.where(
                abs(order[:, i] - np.transpose(norm * o)) > 7. * dev)
            mask[j, i] = 0.0
            dev_new = dev_new + np.sum(
                mask[:, i] * (order[:, i] - np.transpose(norm * o)) ** 2)

    #     # Outputs as desired
    # if (uncert == True) and (im_output == True):
    #     return (sp, unc, im_out, slitfunc_out)
    # elif uncert == True:
    #     return (sp, unc, slitfunc_out)
    # elif im_output == True:
    #     return (sp, im_out, slitfunc_out)
    # else:
    #     return (sp, slitfunc_out)
    
    return dict(
        sp=sp,
        sp_wing0=sp_wing0,
        unc=unc,
        im_out=im_out,
        slitfunc_out=slitfunc_out,
        slitfunc=slitfunc,
        sf_wing0=sf_wing0,
        slitfunc_wing0=slitfunc_wing0/slitfunc_wing0.sum(axis=0),
        sf=sf,
        ssf=ssf,
        sf_med=sf_med,
    )
#    
##%%
#
#from joblib import load, dump
#
## load flat
#flat = load('/home/cham/PycharmProjects/hrs/twodspec/data/flat.dump')
#bias = load('/home/cham/PycharmProjects/hrs/twodspec/data/bias.dump')
#star = load('/home/cham/PycharmProjects/hrs/twodspec/data/star.dump')
#star_sl = load('/home/cham/PycharmProjects/hrs/twodspec/data/star_sl.dump')
#star_q = load('/home/cham/PycharmProjects/hrs/twodspec/data/star1d_q.dump')
#star1d_sw = load('/home/cham/PycharmProjects/hrs/twodspec/data/star_sw.dump')
#flat1d_q = load('/home/cham/PycharmProjects/hrs/twodspec/data/flat1d_q.dump')
#
#
#
#
#
#i_ap=10
#spatial_drift = ap.ap_center_interp[i_ap, chunk_cols[0]:chunk_cols[1]].flatten()
#y1,y2=np.modf(spatial_drift)
#this_chunk = np.zeros((29, 2048), dtype=float)
#for icol in range(2048):
#    this_chunk[:, icol] = star_sl[np.int(y2[icol]-14):np.int(y2[icol]+15), icol]
#
#prdc = pyREDUCE(this_chunk, y1, im_output=True)
#
#figure();
#plot(prdc[0]/star1d_sw[2][49-i_ap]*0.95);
##plot(star_q[10]/star1d_sw[2][49-i_ap])
#plot(star1d_sw[0][49-i_ap]/star1d_sw[2][49-i_ap])
#
#figure();plotcol=2040; plot(this_chunk[:, plotcol]); plot(prdc[1][:, plotcol-0])
#
#
#figure(); imshow(chunk_im_fill0, aspect='auto'); plot(spatial_drift-chunk_my[0,0]+1)
#
#a = np.array([[1,3,5],[7,2,9], [5,5,9]])
#x = np.array([[1],[4],[5]])
#b = np.dot(a, x)
#b = np.array([[9], [2.], [1]])
#
#from scipy.linalg import solve
#x_ = solve(a,b)
#np.dot(a, x_) - b



# -*- coding: utf-8 -*-
"""

Author
------
Bo Zhang

Email
-----
bozhang@nao.cas.cn

Created on
----------
- Fri Nov 25 12:53:24 2016

Modifications
-------------
- re-organized on Tue May 23 21:00:00 2017

Aims
----
- utils for apertures

"""

import numpy as np

from .background import apbackground
from .extract import (get_aperture_section, extract_profile, extract_aperture, extract_all,
                      make_normflat, extract_profile_simple)
from .trace import trace_naive_max


# ###################################################### #
# Here goes the code that will be used after 2017.04.27  #
# ###################################################### #

class Aperture(object):
    """ trace apertures from FLAT or SCI images
    
    The Aperture class defines the framework of Aperture instances.
    However, multiple choices of tracing methods are provided, such as
    1> canny edge detector (for SONG FLAT, implemented)
    2> find local maximum (for HRS, implemented but not integrated)
    
    Other possible methods that could be used to detect apertures:
    1> (Generalized) Hough transformation
    2> Sobel operator?
    3> generalized find local maximum (along an axis but using multiple pixels)
    4> peak local max, from skimage
    5> clustering? sounds infeasible
    6> xcorr. cross-correlation between columns
    
    """
    # image info
    imshape = None
    x = None
    y = None
    mx = None
    my = None

    # trace apertures
    nap = 0  # number of apertures found
    npix = 0
    istraced = False  # True if traced
    method = ""  # method used to trace apertures
    trace_details = None  # raw data maintained during tracing

    # aperture edges, centers and width
    ap_lower = None
    ap_upper = None
    ap_center = None
    ap_width = 0

    # fit apertures & interpolate
    ispolyfitted = False
    polydeg = 0

    # fitted polynomial coefs
    ap_upper_chebcoef = None
    ap_lower_chebcoef = None
    ap_center_chebcoef = None  # center is not fitted in case of canny method

    # interpolated edge & center
    ap_upper_interp = None
    ap_lower_interp = None
    ap_center_interp = None

    def __init__(self, ap_center=np.array([[]]), ap_width=15):
        """ initialize with traces

        Parameters
        ----------
        ap_center:
            aperture center (n_ap x n_pix)
        ap_width:
            aperture width
        """
        ap_center = np.asarray(ap_center, float)
        self.nap, self.npix = ap_center.shape
        self.ap_center = ap_center
        if ap_center is not None:
            self.ap_lower = ap_center - ap_width
            self.ap_upper = ap_center + ap_width
        self.ap_width = ap_width
        return

    @staticmethod
    def trace(flat, method="naive", ap_width=15, polydeg=4, **kwargs):
        """ trace apertures for FLAT with a specified method

        Example
        -------
        >>> from twodspec.aperture import Aperture
        >>> ap = Aperture.trace(flat, method="naive", polydeg=4, sigma=7, maxdev=7, ap_width=15)
        >>> ap = Aperture.trace(flat, method="canny", polydeg=4, sigma=7, maxdev=7, ap_width=15)

        Parameters
        ----------
        flat : ndarray
            FLAT.
        method : str, optional
            {"naive", "canny", "", None}. The default is None.
        ap_width : float, optional
            the half width of the aperture. The default is 15.
        polydeg : int, optional
            The order of polynomial fitting to apertures. The default is 4.
        **kwargs :
            will be passed to trace method.

        Returns
        -------
        ap : Aperture instance
            The Aperture results.

        """
        # assert method is valid
        assert method in {None, "", "canny", "naive"}

        # trace apertures
        print("@Aperture: tracing apertures using [{0}] method".format(method), end="")

        # return null Aperture instance
        if method is None or method == "":
            # initialization
            ap = Aperture()
            # get image info
            ap.get_image_info(flat)
            return ap
        elif method == "canny":
            # 1. canny edge detector
            results, details = trace_canny_col(flat, details=True, verbose=False, sigma=kwargs["sigma"])
            ap = Aperture(ap_center=results["ap_center"], ap_width=ap_width)
            ap.get_image_info(flat)
            ap.trace_details = details
        elif method == "naive":
            # 2. naive max method
            ap_center = trace_naive_max(flat, sigma=kwargs["sigma"], maxdev=kwargs["maxdev"])
            ap = Aperture(ap_center=ap_center, ap_width=ap_width)
            ap.get_image_info(flat)
        else:
            # otherwise
            print("\n@Aperture: invalid method {0}".format(method))
            return Aperture()
        # change status
        ap.method = method
        ap.istraced = True
        # verbose
        print("  >>>  {0} apertures found!".format(ap.nap))

        # polyfit
        if polydeg is not None:
            ap.polyfit(np.int(polydeg))

        return ap

    def get_image_info(self, image):
        """ get image information """
        if not isinstance(image, np.ndarray):
            image = np.array(image)
        self.imshape = image.shape
        self.x = np.arange(self.imshape[1], dtype=int)
        self.y = np.arange(self.imshape[0], dtype=int)
        self.mx, self.my = np.meshgrid(self.x, self.y)
        return

    def polyfit(self, deg=4):
        """ fit using chebyshev polynomial for adopted apertures """
        # interpolated edges
        self.polydeg = deg

        n_row = self.imshape[0]
        nap = self.nap

        ap_col_interp = np.arange(0, n_row, dtype=int)
        ap_upper_interp = []  # interpolated
        ap_lower_interp = []
        ap_center_interp = []
        ap_upper_chebcoef = []  # chebcoef
        ap_lower_chebcoef = []
        ap_center_chebcoef = []
        for i in range(nap):
            # for upper
            ind_fit = self.ap_center[i] > -1
            this_chebcoef = np.polynomial.chebyshev.chebfit(
                self.y[ind_fit], self.ap_upper[i][ind_fit], deg=deg)
            ap_upper_chebcoef.append(this_chebcoef)
            ap_upper_interp.append(
                np.polynomial.chebyshev.chebval(ap_col_interp, this_chebcoef))
            # for lower
            this_chebcoef = np.polynomial.chebyshev.chebfit(
                self.y[ind_fit], self.ap_lower[i][ind_fit], deg=deg)
            ap_lower_chebcoef.append(this_chebcoef)
            ap_lower_interp.append(
                np.polynomial.chebyshev.chebval(ap_col_interp, this_chebcoef))
            # for center
            this_chebcoef = np.polynomial.chebyshev.chebfit(
                self.y[ind_fit], self.ap_center[i][ind_fit], deg=deg)
            ap_center_chebcoef.append(this_chebcoef)
            ap_center_interp.append(
                np.polynomial.chebyshev.chebval(ap_col_interp, this_chebcoef))

        # transform to numpy.array format
        self.ap_upper_interp = np.array(ap_upper_interp)
        self.ap_lower_interp = np.array(ap_lower_interp)
        self.ap_center_interp = np.array(ap_center_interp)
        self.ap_upper_chebcoef = np.array(ap_upper_chebcoef)
        self.ap_lower_chebcoef = np.array(ap_lower_chebcoef)
        self.ap_center_chebcoef = np.array(ap_center_chebcoef)
        # center trace: center is not fitted but averaged from edges
        # self.ap_center_interp = (ap_upper_interp + ap_lower_interp) / 2.

        self.ispolyfitted = True
        return

    def background(self, im, npix_inter=5, q=(40, 5), sigma=(10, 10), kernel_size=(11, 11)):
        """ newly developed on 2017-05-28, with best performance """
        return apbackground(im, self.ap_center, q=q, npix_inter=npix_inter, 
                            sigma=sigma, kernel_size=kernel_size)
    
    def get_aperture_section(self, im, iap):
        """ get an aperture section

        Parameters
        ----------
        im : ndarray
            The target image.
        iap : int
            The aperture index.

        Returns
        -------
        ap_im : ndarray
            DESCRIPTION.
        ap_im_xx : ndarray
            x coordinates.
        ap_im_yy : ndarray
            y coordinates.
        ap_im_xx_cor : ndarray
            x offset from center.

        """
        return get_aperture_section(im, self.ap_center_interp[iap], ap_width=self.ap_width)

    def extract_profile_simple(self, ap_im, ap_im_xx_cor, profile_oversample=10, profile_smoothness=1e-1):

        return extract_profile_simple(ap_im, ap_im_xx_cor, self.ap_width,
                                      profile_oversample=profile_oversample,
                                      profile_smoothness=profile_smoothness)

    def extract_profile(self, ap_im, ap_im_xx_cor, profile_smoothness=1e-2, n_chunks=8,
                        ap_width=15., profile_oversample=10., ndev=4):
        return extract_profile(ap_im, ap_im_xx_cor, profile_smoothness=profile_smoothness, n_chunks=n_chunks,
                               ap_width=ap_width, profile_oversample=profile_oversample, ndev=ndev)

    def make_normflat(self, im, max_dqe=0.04, min_snr=20, smooth_blaze=5, n_chunks=8,
                      profile_oversample=10, profile_smoothness=1e-2, num_sigma_clipping=20, gain=1., ron=0, n_jobs=1):
        """ normalize FLAT
        
        Parameters
        ----------
        im : ndarray
            the target image.
        max_dqe : float, optional
            The max deviation of Quantum Efficiency from 1.0. The default is 0.04.
        min_snr : flaot, optional
            Ignore the region with snr<min_snr. The default is 20.
        smooth_blaze : int, optional
            The smooth kernel width / pixel. The default is 5.
        n_chunks : int, optional
            Split each aperture to n_chunks chunks. The default is 8.
        profile_oversample : int, optional
            Oversampling factor of spatial profile. The default is 10.
        profile_smoothness : float, optional
            The smoothness of the profile. The default is 1e-2.
        num_sigma_clipping : float, optional
            The sigma-clipping value / sigma. The default is 20.
        gain : flaot, optional
            The gain of the image. The default is 1..
        ron : float, optional
            The readout noise. The default is 0.
        n_jobs : int, optional
            The number of processes launched. The default is 1.

        Returns
        -------
        blaze, im_norm : ndarray
            The blaze functions and sensitivity image.

        """
        blaze, im_norm = make_normflat(
            im, self, max_dqe=max_dqe, min_snr=min_snr, smooth_blaze=smooth_blaze, n_chunks=n_chunks,
            profile_oversample=profile_oversample, profile_smoothness=profile_smoothness,
            num_sigma_clipping=num_sigma_clipping, gain=gain, ron=ron, n_jobs=n_jobs)
        return blaze, im_norm

    def extract_all(self, im, n_chunks=8, profile_oversample=10, profile_smoothness=1e-2,
                    num_sigma_clipping=10, gain=1., ron=0, n_jobs=-1,
                    verbose=False, backend="multiprocessing"):
        """ extract all apertures with both simple & profile extraction

        Parameters
        ----------
        im : ndarray
            The target image.
        n_chunks : int, optional
            The number of chunks. The default is 8.
        profile_oversample : int, optional
            The oversampling factor of the profile. The default is 10.
        profile_smoothness : float, optional
            The smoothness of the profile. The default is 1e-2.
        num_sigma_clipping : float, optional
            The sigma clipping threshold. The default is 5..
        gain : float, optional
            The gain of CCD. The default is 1..
        ron : flaot, optional
            The readout noise of CCD. The default is 0.
        n_jobs : int, optional
            The number of processes launched. The default is -1.
        verbose:
            defaults to False
        backend:
            joblib backend

        Returns
        -------
        dict
            a dict sconsisting of many results.

        """

        return extract_all(im, self, n_chunks=n_chunks, profile_oversample=profile_oversample,
                           profile_smoothness=profile_smoothness, num_sigma_clipping=num_sigma_clipping,
                           gain=gain, ron=ron, n_jobs=n_jobs, verbose=verbose, backend=backend)

    def extract_aperture(self, iap, im, n_chunks=8, profile_oversample=10, profile_smoothness=1e-2,
                         num_sigma_clipping=5., gain=1., ron=0):
        """ Extract one aperture

        Parameters
        ----------
        iap : int
            Extract the iap th aperture.
        im : ndarray
            The target image.
        n_chunks : int, optional
            The number of chunks. The default is 8.
        profile_oversample : int, optional
            The oversampling factor of the profile. The default is 10.
        profile_smoothness : float, optional
            The smoothness of the profile. The default is 1e-2.
        num_sigma_clipping : float, optional
            The sigma clipping threshold. The default is 5..
        gain : float, optional
            The gain of CCD. The default is 1..
        ron : flaot, optional
            The readout noise of CCD. The default is 0.

        Returns
        -------
        dict
            # ----- combined extraction -----
            spec_extr=spec_extr,
            err_extr=err_extr,
            # ----- first extraction -----
            spec_extr1=spec_extr1,
            err_extr1=err_extr1,
            # ----- second extraction (after sigma-clipping) -----
            spec_extr2=spec_extr2,
            err_extr2=err_extr2,
            # ----- inconsistent pixels between 1&2, True if difference > 3sigma -----
            mask_extr=mask_extr,
            # ----- simple extraction -----
            spec_sum=ap_im.sum(axis=1),
            err_sum=ap_im_err.sum(axis=1),
            # ----- reconstructed profile -----
            prof_recon=prof_recon,
            # ----- reconstructed aperture -----
            ap_im=ap_im,
            ap_im_recon=ap_im_recon,
            ap_im_recon1=ap_im_recon1,
            ap_im_recon2=ap_im_recon2,
            # ----- aperture coordinates -----
            ap_im_xx=ap_im_xx,
            ap_im_xx_cor=ap_im_xx_cor,
            ap_im_yy=ap_im_yy,.

        """
        return extract_aperture(im, self.ap_center_interp[iap], n_chunks=n_chunks, ap_width=self.ap_width,
                                profile_oversample=profile_oversample, profile_smoothness=profile_smoothness,
                                num_sigma_clipping=num_sigma_clipping, gain=gain, ron=ron)


def test_aperture():
    from astropy.io import fits
    import matplotlib.pyplot as plt
    plt.rcParams.update({"font.size": 15})
    flat = fits.getdata("/Users/cham/projects/song/star_spec/20191105/night/ext/masterflat_20191105_slit5.fits")

    ap = Aperture.trace(flat, method="naive", polydeg=4, sigma=7, maxdev=7, ap_width=15)
    ap = Aperture.trace(flat, method="canny", polydeg=4, sigma=7, maxdev=7, ap_width=15)
    return ap


if __name__ == "__main__":
    test_aperture()



# -*- coding: utf-8 -*-
"""

Author
------
Bo Zhang

Email
-----
bozhang@nao.cas.cn

Created on
----------
- Mon Nov 28 15:00:00 2016

Modifications
-------------
-

Aims
----
- wavelength calibration module
    all operations are based on extracted ThAr (1d)

"""

import itertools

import numpy as np
from astropy.io import fits
from joblib import Parallel, delayed
from scipy.interpolate import splrep, splev
from scipy.optimize import curve_fit, leastsq


# ############################## #
#      fix thar spectra
# ############################## #

def thar1d_fix(thar1d, conv_len=20, sat_count=50000):
    """ fix thar image, particularly for negative & saturated pixels

    Parameters
    ----------
    conv_len:
        convolved length
    sat_count:
        saturated count / max count

    Return
    ------
    thar1d_fixed:
        fixed THAR image

    """
    # ind_sat_conv
    ind_sat = thar1d >= sat_count
    ind_sat_conv = np.zeros_like(ind_sat)
    x_sat, y_sat = np.where(ind_sat)
    for i in range(len(x_sat)):
        ind_sat_conv[x_sat[i], y_sat[i] - conv_len:y_sat[i] + conv_len] = 1
    # ind_neg
    ind_neg = thar1d < 0.
    # combine to ind_bad
    ind_bad = np.logical_or(ind_neg, ind_sat_conv)
    thar1d_fixed = np.where(ind_bad, 0, thar1d)

    return thar1d_fixed


# ############################## #
#  2d correlation for 1d thar
# ############################## #

def thar1d_corr2d(thar1d_fixed, thar_temp, x_shiftmax=20, y_shiftmax=5,
                  verbose=False):
    """ determine the shift of *thar1d_fixed* relative to *thar_temp*

    Parameters
    ----------
    thar1d_fixed:
        image whose shift is to be determined
    thar_temp:
        reference image
    xtrim: tuple
        (xmin, xmax)
    ytrim: tuple
        (ymin, ymax)
    x_shiftmax: int
        the max y shift for correlation
    y_shiftmax: int
        the max y shift for correlation

    Returns
    -------
    (x_shift, y_shift), corr2d

    """

    # default trim region is the center 1/4 area
    xtrim = (thar1d_fixed.shape[1] * np.array([.2, .8])).astype(int)
    ytrim = (thar1d_fixed.shape[0] * np.array([.2, .8])).astype(int)

    # in case that the input data are int
    thar1d_fixed = np.array(thar1d_fixed).astype(float)
    thar_temp = np.array(thar_temp).astype(float)

    if verbose:
        print("@Cham: computing 2D cross-correlation...")

    # initialize result
    corr2d = np.zeros((1 + 2 * y_shiftmax, 1 + 2 * x_shiftmax))
    # make slice
    xslice_temp = slice(*xtrim)
    yslice_temp = slice(*ytrim)
    # 2D correlation
    for xofst in range(-x_shiftmax, x_shiftmax + 1, 1):
        for yofst in range(-y_shiftmax, y_shiftmax + 1, 1):
            xslice = slice(xtrim[0] + xofst, xtrim[1] + xofst, 1)
            yslice = slice(ytrim[0] + yofst, ytrim[1] + yofst, 1)
            corr2d[yofst + y_shiftmax, xofst + x_shiftmax] = \
                np.mean(thar1d_fixed[yslice, xslice] *
                        thar_temp[yslice_temp, xslice_temp])
    # select maximum value
    y_, x_ = np.where(corr2d == np.max(corr2d))
    x_shift = np.int(x_shiftmax - x_)  # reverse sign
    y_shift = np.int(y_shiftmax - y_)

    return (x_shift, y_shift), corr2d


# ############################## #
#      shift wave & order
# ############################## #

def interpolate_wavelength(w, shift, thar_temp, thar1d_fixed):
    """ interpolate given a regerence wavelength solution and shift """
    xshift, yshift = shift

    xcoord = np.arange(w.shape[1])
    ycoord = np.arange(w.shape[0])

    # for X, no difference, for Y, orders are different
    xcoord_xshift = xcoord + xshift
    ycoord_yshift = np.arange(
        w.shape[0] + thar1d_fixed.shape[0] - thar_temp.shape[0]) + yshift
    # shfit X
    w_x = np.zeros(w.shape)
    for i in range(w.shape[0]):
        s = splrep(xcoord, w[i], k=3, s=0)
        w_x[i] = splev(xcoord_xshift, s)

    # shift Y
    w_x_y = np.zeros((thar1d_fixed.shape[0], w.shape[1]))
    for j in range(w.shape[1]):
        s = splrep(ycoord, w_x[:, j], k=3, s=0)
        w_x_y[:, j] = splev(ycoord_yshift, s)

    return w_x_y


def interpolate_order(order_temp, shift, thar1d_fixed):
    xshift, yshift = shift

    # for X, no difference, for Y, orders are different
    ycoord_yshift = np.arange(thar1d_fixed.shape[0]) + \
                    order_temp[0, 0] + yshift
    order_interp = np.repeat(ycoord_yshift.reshape(-1, 1),
                             thar1d_fixed.shape[1], axis=1)

    return order_interp


# ############################## #
#      load temp thar spectra
# ############################## #

def load_thar_temp(thar_temp_path):
    hl = fits.open(thar_temp_path)
    wave = hl['wave'].data
    thar = hl['thar'].data
    order = hl['order'].data
    return wave, thar, order


# ############################## #
#      refine thar positions
# ############################## #

def refine_thar_positions(wave_init, order_init, thar1d_fixed, thar_list,
                          fit_width=5., lc_tol=5., k=3, n_jobs=10, verbose=10):
    """ refine ThAr positions """
    print("@TWODSPEC: refine ThAr positions ...")

    # refine thar positions for each order
    r = Parallel(n_jobs=n_jobs, verbose=verbose, batch_size=1)(
        delayed(refine_thar_positions_order)(
            wave_init[i_order],
            np.arange(wave_init.shape[1]),
            thar1d_fixed[i_order],
            thar_list[(thar_list > np.min(wave_init[i_order]) + 1.) * (
                thar_list < np.max(wave_init[i_order]) - 1.)],
            order_init[i_order, 0],
            fit_width=fit_width, lc_tol=lc_tol, k=k
        ) for i_order in range(wave_init.shape[0]))

    # remove all null values
    null_value = None
    for i in range(r.count(null_value)):
        r.remove(null_value)
        # print(len(r))

    # collect data
    lc_coord = np.array(np.hstack([_[0] for _ in r]))
    lc_order = np.array(np.hstack([_[1] for _ in r]))
    lc_thar = np.array(np.hstack([_[2] for _ in r]))
    popt = np.array(np.vstack([_[3] for _ in r]))
    pcov = np.array(np.vstack([_[4] for _ in r]))

    return lc_coord, lc_order, lc_thar, popt, pcov


def refine_thar_positions_order(this_wave_init, this_xcoord, this_thar,
                                this_thar_list, this_order, fit_width=5.,
                                lc_tol=5., k=3):
    if len(this_thar_list) == 0:
        return None

    popt_list = []
    pcov_list = []

    # refine all thar positions in this order
    for i_thar_line, each_thar_line in enumerate(this_thar_list):

        # cut local spectrum
        ind_local = (this_wave_init > each_thar_line - fit_width) * (
            this_wave_init < each_thar_line + fit_width)

        # set bounds
        p0 = (0., 1E5, each_thar_line, 0.1)
        bounds = ((-1., 0., each_thar_line - lc_tol, 0.01),
                  (+np.inf, np.inf, each_thar_line + lc_tol, 2.))

        try:
            popt, pcov = curve_fit(gauss_poly0, this_wave_init[ind_local],
                                   this_thar[ind_local], p0=p0, bounds=bounds)
            pcov = np.diagonal(pcov)
        except RuntimeError:
            popt = np.ones_like(p0) * np.nan
            pcov = np.ones((len(p0),)) * np.nan

        popt_list.append(popt)
        pcov_list.append(pcov)

    # interpolation for X corrdinates
    if np.all(np.diff(this_wave_init) >= 0):
        tck = splrep(this_wave_init, this_xcoord, k=k)
        # print(this_wave_init, this_xcoord, k)
    elif np.all(np.diff(this_wave_init) <= 0):
        tck = splrep(this_wave_init[::-1], this_xcoord[::-1], k=k)
        print("@SONG: wavelength inversed!")
    else:
        raise (ValueError("@Cham: error occurs in interpolation!"))

    # lccov_list = np.array(lccov_list)
    popt_list = np.array(popt_list)
    pcov_list = np.array(pcov_list)

    lc_coord = splev(popt_list[:, 2], tck)
    lc_order = np.ones_like(lc_coord) * this_order

    return lc_coord, lc_order, this_thar_list, popt_list, pcov_list


# ############################## #
#      2D surface fit
# ############################## #

def polyval2d(x, y, coefs, orders=None):
    if orders is None:
        orderx, ordery = coefs.shape
    else:
        orderx, ordery = orders

    ij = itertools.product(range(orderx + 1), range(ordery + 1))
    z = np.zeros_like(x)
    for a, (i, j) in zip(coefs.flatten(), ij):
        if i + j < np.max((orderx, ordery)):
            #        print a,i,j
            z += a * x ** i * y ** j
    return z


def gauss_poly1(x, p0, p1, a, b, c):
    return p0 + p1 * x + a/np.sqrt(2.*np.pi)/c * np.exp(-0.5*((x - b) / c) ** 2.)


def gauss_poly0(x, p0, a, b, c):
    return p0 + a/np.sqrt(2.*np.pi)/c * np.exp(-0.5 * ((x - b) / c) ** 2.)


def gauss(x, a, b, c):
    return a/np.sqrt(2.*np.pi)/c * np.exp(-0.5 * ((x - b) / c) ** 2.)


def residual_chi2(coefs, x, y, z, w, poly_order):
    return np.nansum(residual(coefs, x, y, z, w, poly_order) ** 2.)


def residual_lar(coefs, x, y, z, w, poly_order):
    fitted = polyval2d(x, y, coefs, poly_order)
    return np.sqrt(np.abs((fitted - z) * w))


def residual(coefs, x, y, z, w, poly_order):
    fitted = polyval2d(x, y, coefs, poly_order)
    return (fitted - z) * w


# ############################## #
#      standardization
# ############################## #

def standardize_inverse(x, xmean, xstd):
    return (x * xstd) + xmean


def standardize(x):
    return (x - np.nanmean(x)) / np.nanstd(x), np.nanmean(x), np.nanstd(x)


# ############################## #
#      fit grating equation
# ############################## #

def fit_grating_equation(lc_coord, lc_order, lc_thar, popt, pcov, ind_good_thar0=None,
                         poly_order=(3, 5), max_dev_threshold=100, n_iter=400, lar=False,
                         nl_eachorder=10):
    # pick good thar lines
    if ind_good_thar0 is None:
        ind_good_thar0 = np.ones_like(lc_coord, dtype=bool)
    try:
        ind_good_thar = ind_good_thar0 * np.isfinite(lc_coord) * \
                        (popt[:, 3] < 1.0) * (pcov[:, 2] < 1.0)
    except:
        ind_good_thar = np.ones_like(lc_coord, dtype=bool)

    lc_coord = lc_coord[ind_good_thar]
    lc_order = lc_order[ind_good_thar]
    lc_thar = lc_thar[ind_good_thar]

    # standardization
    lc_coord_s, lc_coord_mean, lc_coord_std = standardize(lc_coord)
    lc_order_s, lc_order_mean, lc_order_std = standardize(lc_order)
    ml_s, ml_mean, ml_std = standardize(lc_thar * lc_order)
    # scaler
    scaler_coord = lc_coord_mean, lc_coord_std
    scaler_order = lc_order_mean, lc_order_std
    scaler_ml = ml_mean, ml_std

    # weight
    weight = popt[ind_good_thar, 1] / popt[ind_good_thar, 0]
    weight = weight > 0

    # fit surface
    x0 = np.zeros(poly_order)
    #print(lc_coord_s, lc_order_s, ml_s, weight, poly_order)
    x0, ier = leastsq(residual, x0,
                      args=(lc_coord_s, lc_order_s, ml_s, weight, poly_order))
    #print(x0, ier)

    # iter
    if n_iter > 0:
        n_loop = 0
        while n_loop < n_iter:
            n_loop += 1
            if lar:
                x_mini_lsq, ier = leastsq(residual_lar, x0, args=(
                    lc_coord_s, lc_order_s, ml_s, weight, poly_order))
            else:
                x_mini_lsq, ier = leastsq(residual, x0, args=(
                    lc_coord_s, lc_order_s, ml_s, weight, poly_order))
            fitted = polyval2d(lc_coord_s, lc_order_s, x_mini_lsq, poly_order)
            fitted_wave = standardize_inverse(fitted, ml_mean, ml_std) / lc_order
            fitted_wave_diff = fitted_wave - lc_thar

            wave_dev = np.abs(fitted_wave_diff - np.nanmedian(fitted_wave_diff))
            lc_order_s_unique, ind_inverse = np.unique(lc_order_s, return_inverse=True)
            lc_order_s_unique_left = np.zeros_like(lc_order_s_unique, dtype=int)
            for i, _ in enumerate(lc_order_s_unique):
                lc_order_s_unique_left[i] = np.sum(weight[lc_order_s == _] > 0)
            lc_order_s_left = lc_order_s_unique_left[ind_inverse]
            # print(lc_order_s_left)
            # print(wave_dev)
            # print(lc_order_s_left)

            possible_outlier = wave_dev * (wave_dev > max_dev_threshold) * (lc_order_s_left > nl_eachorder)
            if np.any(possible_outlier > 0):
                ind_max_dev = np.nanargmax(possible_outlier)
                weight[ind_max_dev] = 0.
                ml_s[ind_max_dev] = np.nan
                lc_thar[ind_max_dev] = np.nan
                print("@Cham: [n_loop = %s] max_dev = %s" % (
                    n_loop, fitted_wave_diff[ind_max_dev]))
            else:
                if n_loop == 0:
                    print("@Cham: no points cut in iterarions ...")
                break
    else:
        if lar:
            x_mini_lsq, ier = leastsq(residual_lar, x0, args=(
                lc_coord_s, lc_order_s, ml_s, weight, poly_order))
        else:
            x_mini_lsq, ier = leastsq(residual, x0, args=(
                lc_coord_s, lc_order_s, ml_s, weight, poly_order))
        fitted = polyval2d(lc_coord_s, lc_order_s, x_mini_lsq, poly_order)
        fitted_wave = standardize_inverse(fitted, ml_mean, ml_std) / lc_order
        fitted_wave_diff = fitted_wave - lc_thar

        ind_kick = np.abs(fitted_wave_diff - np.nanmedian(fitted_wave_diff)) > max_dev_threshold
        weight[ind_kick] = 0.
        ml_s[ind_kick] = np.nan
        lc_thar[ind_kick] = np.nan
        x_mini_lsq, ier = leastsq(residual_lar, x0, args=(
            lc_coord_s, lc_order_s, ml_s, weight, poly_order))

    ind_good_thar = np.where(ind_good_thar)[0][np.isfinite(lc_thar)]
    print("@SONG: RMS = {0} | n_points = {1} ".format(
        np.sqrt(np.nanmean(np.square(fitted_wave - lc_thar))),
        np.sum(np.isfinite(lc_thar)))
    )

    return x_mini_lsq, ind_good_thar, scaler_coord, scaler_order, scaler_ml


def grating_equation_predict(grid_coord, grid_order, x_mini_lsq, poly_order,
                             scaler_coord, scaler_order, scaler_ml):
    sgrid_coord = (grid_coord - scaler_coord[0]) / scaler_coord[1]
    sgrid_order = (grid_order - scaler_order[0]) / scaler_order[1]

    sgrid_fitted = polyval2d(sgrid_coord.flatten(), sgrid_order.flatten(),
                             x_mini_lsq, poly_order)
    sgrid_fitted_wave = (sgrid_fitted * scaler_ml[1] + scaler_ml[0]) / \
                        grid_order.flatten()
    sgrid_fitted_wave = sgrid_fitted_wave.reshape(grid_coord.shape)
    return sgrid_fitted_wave


def polyfit_reject(x, y, deg=1, w=None, epsilon=0.002, n_reserve=-1):

    x, y = np.array(x), np.array(y)
    ind_good = np.ones_like(x, bool)
    x_, y_ = x[ind_good], y[ind_good]

    if w is None:
        w = np.ones_like(x, float)
    w_ = np.array(w)[ind_good]

    n_good = np.sum(ind_good)
    if isinstance(n_reserve, float):
        # fraction
        assert 0 < n_reserve < 1
        n_reserve = np.int(n_good * n_reserve)

    # set up the iteration
    ind_reserved = np.ones_like(x_, bool)
    while np.sum(ind_reserved) > n_reserve:
        # p0 = np.polyfit(x_[ind_reserved], y_[ind_reserved], deg, w=w_[ind_reserved])
        p0, ier = leastsq(polyfit_costfun_lar, np.zeros((deg+1,), float), args=(x_[ind_reserved], y_[ind_reserved], w_[ind_reserved]))

        y_res = y_ - np.polyval(p0, x_)
        y_res = np.where(w_ > 0, y_res, np.nan)
        # y_res_std = np.nanstd(y_res)
        # y_res_med = np.nanmedian(y_res)
        # y_res_scaled = (y_res>0) * (y_res/y_res_std/sigma_reject[0]) + (y_res<=0) * (y_res/y_res_std/sigma_reject[1])
        ind_bad = np.abs(y_res) > epsilon

        if np.any(ind_bad[ind_reserved]):
            # print(np.sum(ind_bad[ind_reserved]))
            i_rejected = np.nanargmax(np.abs(y_res))
            # print(i_rejected, np.abs(y_res)[i_rejected])
            ind_reserved[i_rejected] = False
            w_[i_rejected] = 0
            # print("@SONG: {0} points left".format(np.sum(ind_reserved)))
        else:
            break
    p = np.polyfit(x_[ind_reserved], y_[ind_reserved], deg, w=w_[ind_reserved])
    return p, w_>0


def polyfit_costfun_lar(p, x, y, w):
    return np.sqrt(np.abs(np.polyval(p, x) - y)) * w


def polyfit_costfun(p, x, y, w):
    return (np.polyval(p, x) - y) * w


def clean_thar_polyfit1d_reject(lc_coord, lc_order, lc_thar, popt,
                                ind_good0=None, deg=1, w=None,
                                epsilon=0.004, n_reserve=3):
    if ind_good0 is not None:
        lc_coord[~ind_good0] = np.nan
    print("@TWODSPEC: number of finite line centers: {}".format(
        np.sum(np.isfinite(lc_coord))))

    sub_good = np.zeros((0,), int)
    for i_order in np.unique(lc_order)[:]:
        this_ind = lc_order == i_order
        this_sub = np.where(this_ind)[0]
        this_lc_coord = lc_coord[this_ind]
        # this_lc_order = lc_order[this_ind]
        this_lc_thar = lc_thar[this_ind]

        x, y, z = this_lc_coord, popt[this_ind, 2] - this_lc_thar, this_lc_thar
        ind_use = np.isfinite(x) * np.isfinite(y)
        x, y, z = x[ind_use], y[ind_use], z[ind_use]

        p, ind_reserved = polyfit_reject(x, y, deg, w, epsilon=epsilon, n_reserve=n_reserve)
        sub_good = np.hstack((sub_good, this_sub[ind_use][ind_reserved]))
        # print(len(this_sub[ind_use][ind_reserved]))

    ind_good = np.zeros_like(lc_coord, bool)
    ind_good[sub_good] = True
    return ind_good


# -*- coding: utf-8 -*-
"""

Author
------
Bo Zhang

Email
-----
bozhang@nao.cas.cn

Created on
----------
- Fri Nov 25 12:53:24 2016

Modifications
-------------
- re-organized on Tue May 23 21:00:00 2017

Aims
----
- utils for apertures

"""

import warnings
from copy import copy

import numpy as np
from scipy.interpolate import interp1d
from scipy.optimize import leastsq
from scipy.signal import medfilt2d, medfilt
from scipy.stats import binned_statistic

from .background import apbackground
from .trace import trace_canny_col, trace_naive_max
from .deprecated.aprec import AprecList, Aprec  # deprecated IRAF interface


# ###################################################### #
# Here goes the code that will be used after 2017.04.27  #
# ###################################################### #

class Aperture(object):
    """ trace apertures from FLAT or SCI images
    
    The Aperture class defines the framework of Aperture instances.
    However, multiple choices of tracing methods are provided, such as
    1> canny edge detector (for SONG FLAT, implemented)
    2> find local maximum (for HRS, implemented but not integrated)
    
    Other possible methods that could be used to detect apertures:
    1> (Generalized) Hough transformation
    2> Sobel operator?
    3> generalized find local maximum (along an axis but using multiple pixels)
    4> peak local max, from skimage
    5> clustering? sounds infeasible
    6> xcorr. cross-correlation between columns
    
    """
    # image info
    imshape = None
    x = None
    y = None
    mx = None
    my = None

    # trace apertures
    nap = 0  # number of apertures found
    npix = 0
    istraced = False  # True if traced
    method = ""  # method used to trace apertures
    trace_details = None  # raw data maintained during tracing

    # aperture edges, centers and width
    ap_lower = None
    ap_upper = None
    ap_center = None
    ap_width = 0

    # fit apertures & interpolate
    ispolyfitted = False
    polydeg = 0

    # fitted polynomial coefs
    ap_upper_chebcoef = None
    ap_lower_chebcoef = None
    ap_center_chebcoef = None  # center is not fitted in case of canny method

    # interpolated edge & center
    ap_upper_interp = None
    ap_lower_interp = None
    ap_center_interp = None

    def __init__(self, ap_center=np.array([[]]), ap_width=15):
        """ initialize with traces

        Parameters
        ----------
        ap_center:
            aperture center (n_ap x n_pix)
        ap_width:
            aperture width
        """
        ap_center = np.asarray(ap_center, float)
        self.nap, self.npix = ap_center.shape
        self.ap_center = ap_center
        if ap_center is not None:
            self.ap_lower = ap_center - ap_width
            self.ap_upper = ap_center + ap_width
        return

    @staticmethod
    def trace(flat, method="naive", ap_width=15, polydeg=4, **kwargs):
        """ trace apertures for FLAT with a specified method

        Example
        -------
        >>> from twodspec.aperture import Aperture
        >>> ap = Aperture.trace(flat, method="naive", polydeg=4, sigma=7, maxdev=7, ap_width=15)
        >>> ap = Aperture.trace(flat, method="canny", polydeg=4, sigma=7, maxdev=7, ap_width=15)

        Parameters
        ----------
        flat : ndarray
            FLAT.
        method : str, optional
            {"naive", "canny", "", None}. The default is None.
        ap_width : float, optional
            the half width of the aperture. The default is 15.
        polydeg : int, optional
            The order of polynomial fitting to apertures. The default is 4.
        **kwargs :
            will be passed to trace method.

        Returns
        -------
        ap : Aperture instance
            The Aperture results.

        """
        # assert method is valid
        assert method in {None, "", "canny", "naive"}

        # trace apertures
        print("@Aperture: tracing apertures using [{0}] method".format(method), end="")

        # return null Aperture instance
        if method is None or method == "":
            # initialization
            ap = Aperture()
            # get image info
            ap.get_image_info(flat)
            return ap
        elif method == "canny":
            # 1. canny edge detector
            results, details = trace_canny_col(flat, details=True, verbose=False, sigma=kwargs["sigma"])
            ap = Aperture(ap_center=results["ap_center"], ap_width=ap_width)
            ap.get_image_info(flat)
            ap.trace_details = details
        elif method == "naive":
            # 2. naive max method
            ap_center = trace_naive_max(flat, sigma=kwargs["sigma"], maxdev=kwargs["maxdev"])
            ap = Aperture(ap_center=ap_center, ap_width=ap_width)
            ap.get_image_info(flat)
        else:
            # otherwise
            print("\n@Aperture: invalid method {0}".format(method))
            return Aperture()
        # change status
        ap.method = method
        ap.istraced = True
        # verbose
        print("  >>>  {0} apertures found!".format(ap.nap))

        # polyfit
        if polydeg is not None:
            ap.polyfit(np.int(polydeg))

        return ap

    def get_image_info(self, image):
        """ get image information """
        if not isinstance(image, np.ndarray):
            image = np.array(image)
        self.imshape = image.shape
        self.x = np.arange(self.imshape[1], dtype=int)
        self.y = np.arange(self.imshape[0], dtype=int)
        self.mx, self.my = np.meshgrid(self.x, self.y)
        return

    def polyfit(self, deg=4):
        """ fit using chebyshev polynomial for adopted apertures """
        # interpolated edges
        self.polydeg = deg

        n_row = self.imshape[0]
        nap = self.nap

        ap_col_interp = np.arange(0, n_row, dtype=int)
        ap_upper_interp = []  # interpolated
        ap_lower_interp = []
        ap_center_interp = []
        ap_upper_chebcoef = []  # chebcoef
        ap_lower_chebcoef = []
        ap_center_chebcoef = []
        for i in range(nap):
            # for upper
            this_chebcoef = np.polynomial.chebyshev.chebfit(
                self.y, self.ap_upper[i], deg=deg)
            ap_upper_chebcoef.append(this_chebcoef)
            ap_upper_interp.append(
                np.polynomial.chebyshev.chebval(ap_col_interp, this_chebcoef))
            # for lower
            this_chebcoef = np.polynomial.chebyshev.chebfit(
                self.y, self.ap_lower[i], deg=deg)
            ap_lower_chebcoef.append(this_chebcoef)
            ap_lower_interp.append(
                np.polynomial.chebyshev.chebval(ap_col_interp, this_chebcoef))
            # for center
            this_chebcoef = np.polynomial.chebyshev.chebfit(
                self.y, self.ap_center[i], deg=deg)
            ap_center_chebcoef.append(this_chebcoef)
            ap_center_interp.append(
                np.polynomial.chebyshev.chebval(ap_col_interp, this_chebcoef))

        # transform to numpy.array format
        self.ap_upper_interp = np.array(ap_upper_interp)
        self.ap_lower_interp = np.array(ap_lower_interp)
        self.ap_center_interp = np.array(ap_center_interp)
        self.ap_upper_chebcoef = np.array(ap_upper_chebcoef)
        self.ap_lower_chebcoef = np.array(ap_lower_chebcoef)
        self.ap_center_chebcoef = np.array(ap_center_chebcoef)
        # center trace: center is not fitted but averaged from edges
        # self.ap_center_interp = (ap_upper_interp + ap_lower_interp) / 2.

        self.ispolyfitted = True
        return

    def background(self, im, npix_inter=5, q=(40, 5), sigma=(10, 10), kernel_size=(11, 11)):
        """ newly developed on 2017-05-28, with best performance """
        return apbackground(im, self.ap_center, q=q,
                            npix_inter=npix_inter, sigma=sigma,
                            kernel_size=kernel_size)

    # def to_apreclist(self, comment=None, image="flat.fits", axis=2,
    #                  disp_range=(1, 2048), disp_center=1024,
    #                  function="chebyshev", order=5, width=(12, 12)):
    #     """ convert to AprecList instance """
    #
    #     if function == "chebyshev":
    #         function = 1
    #
    #     if isinstance(width, int) or isinstance(width, float):
    #         width = (np.abs(width), np.abs(width))
    #
    #     if axis == 1:
    #         raise(NotImplementedError("Not implemented for axis = 1!"))
    #
    #         al = AprecList()
    #         x = self.x + 1  # difference between 1-indexed and 0-indexed array
    #
    #         ar_list = []
    #         for i in range(self.n_ap):
    #             y = self.ap_center[i] + 1
    #
    #             ar = Aprec()
    #             ar.image = image
    #             ar.apID = i + 1
    #             ar.beamID = i + 1
    #             ar.low = (-width[0], disp_range[0] - disp_center)
    #             ar.high = (width[1], disp_range[1] - disp_center)
    #             ar.axis = axis
    #
    #             scaler = Scaler(x, (-1, 1))
    #             # the curve range
    #             ar.cvrec.range = scaler.scale_back([-1, 1])
    #             # fit scaled x & y
    #             x_scaled = scaler.scale(x)
    #             this_chebcoef = np.polynomial.chebyshev.chebfit(
    #                 x_scaled, y, deg=order - 1)
    #             this_y_center = np.polynomial.chebyshev.chebval(
    #                 scaler.scale(disp_center), this_chebcoef)
    #
    #             ar.center = this_y_center, disp_center
    #             this_chebcoef[0] -= this_y_center
    #             ar.cvrec.coefs = this_chebcoef
    #             ar.cvrec.function = function
    #             ar.cvrec.order = order
    #
    #             ar_list.append(ar)
    #
    #         al.data = ar_list
    #         return al
    #     else:
    #
    #         al = AprecList()
    #         x = self.y   # difference between 1-indexed and 0-indexed array
    #
    #         ar_list = []
    #         for i in range(self.n_ap):
    #             # since the rot90 operation:
    #             y = self.ap_center[i]
    #
    #             ar = Aprec()
    #             ar.image = image
    #             ar.apID = i + 1
    #             ar.beamID = i + 1
    #             ar.low = (-width[0], disp_range[0] - disp_center)
    #             ar.high = (width[1], disp_range[1] - disp_center)
    #             ar.axis = axis
    #
    #             scaler = Scaler(x, (-1, 1))
    #             # the curve range
    #             ar.cvrec.range = scaler.scale_back([-1, 1])
    #             # fit scaled x & y
    #             x_scaled = scaler.scale(x)
    #             this_chebcoef = np.polynomial.chebyshev.chebfit(
    #                 x_scaled, y, deg=order - 1)
    #             this_y_center = np.polynomial.chebyshev.chebval(
    #                 scaler.scale(disp_center), this_chebcoef)
    #
    #             ar.center = this_y_center, disp_center
    #             this_chebcoef[0] -= this_y_center
    #             ar.cvrec.coefs = this_chebcoef
    #             ar.cvrec.function = function
    #             ar.cvrec.order = order
    #
    #             ar_list.append(ar)
    #
    #         al.data = ar_list
    #         return al

    # def ap_region(self, i_ap=0, n_extend=0):
    #     """ return aperture region & weight for a specified aperture """
    #
    #     this_aperture_weight = np.zeros_like(self.mx, dtype=float)
    #     this_aperture_condition = np.logical_and(
    #         self.my > np.floor(self.ap_upper_interp[i_ap] - n_extend),
    #         self.my < np.ceil(self.ap_lower_interp[i_ap] + n_extend))
    #     min_disance_to_edge = np.min(np.abs(np.array(
    #         (self.my - self.ap_upper_interp[i_ap] + n_extend,
    #          self.my - self.ap_lower_interp[i_ap] - n_extend))), axis=0)
    #     this_aperture_weight = np.where(
    #         this_aperture_condition, min_disance_to_edge,
    #         this_aperture_weight)
    #     this_aperture_weight = np.where(
    #         this_aperture_weight > 1, 1.0, this_aperture_weight)
    #
    #     return this_aperture_condition, this_aperture_weight
    #
    # def ap_region_all(self, n_extend=0):
    #     """ return aperture region & weight for all apertures """
    #
    #     ap_region_all = np.zeros((self.nap, *self.imshape), dtype=bool)
    #     ap_weight_all = np.zeros((self.nap, *self.imshape), dtype=float)
    #
    #     for i in range(self.nap):
    #         ap_region_all[i], ap_weight_all[i] = \
    #             self.ap_region(i, n_extend=n_extend)
    #
    #     return ap_region_all, ap_weight_all
    #
    # def ap_region_chunk(self, i_ap=0, chunk_cols=(0, 2048), n_extend=0, n_cushion=3, scheme='global'):
    #     """  return chunk slice, data and mask
    #
    #     Parameters
    #     ----------
    #     i_ap: int
    #         the aperture serial number
    #     chunk_cols: tuple
    #         (start column, stop column)
    #     n_extend: int
    #         aperture extension
    #     n_cushion: int
    #         cushion for cutting a chunk, could be filled with specified value
    #     scheme: string
    #         if local, return a local cut
    #         if global, upper & lower limit dtermined using the whole aperture
    #
    #     Returns
    #     -------
    #
    #     """
    #
    #     n_cushion = np.int(n_cushion)
    #
    #     # column slice
    #     slice_col = slice(np.int(chunk_cols[0]), np.int(chunk_cols[1]), 1)
    #
    #     # row slice
    #     if scheme == 'local':
    #         lower_limit = np.int(
    #             np.ceil(self.ap_lower_interp[i_ap, slice_col].max()))
    #         upper_limit = np.int(
    #             np.floor(self.ap_upper_interp[i_ap, slice_col].min()))
    #     elif scheme == "global":
    #         lower_limit = np.int(np.ceil(self.ap_lower_interp[i_ap, :].max()))
    #         upper_limit = np.int(np.floor(self.ap_upper_interp[i_ap, :].min()))
    #     else:
    #         raise (ValueError("@Aperture: Invalid scheme: {0}".format(scheme)))
    #     slice_row = slice(np.int(upper_limit - n_cushion - n_extend),
    #                       np.int(lower_limit + 1 + n_cushion + n_extend), 1)
    #
    #     # cushion_mask
    #     this_aperture_condition, this_aperture_weight = \
    #         self.ap_region(i_ap=i_ap, n_extend=n_extend)
    #     chunk_mask = this_aperture_condition[slice_row, slice_col]
    #     chunk_weight = this_aperture_weight[slice_row, slice_col]
    #
    #     return dict(
    #         slice_=(slice_row, slice_col),
    #         slice_row=slice_row,
    #         slice_col=slice_col,
    #         mask=chunk_mask,
    #         weight=chunk_weight
    #     )
    #
    # def ap_region_chunk_rdc(self, i_ap=0, cols=(0, 512), widths=(15, 15)):
    #     """ this is to get the chunk for pyREDUCE """
    #     y1, y2 = np.modf(self.ap_center_interp[i_ap][cols[0]:cols[1]])
    #     nrow = np.int(widths[1] + widths[0])
    #     ncol = np.int(cols[1] - cols[0])
    #     slice1 = np.repeat(
    #         np.arange(cols[0], cols[1], 1, dtype=int).reshape(1, -1),
    #         nrow, axis=0)
    #     slice0 = np.repeat(
    #         np.arange(-widths[0], widths[1], 1, dtype=int).reshape(-1, 1),
    #         ncol, axis=1) + y2.astype(int)
    #     return slice0, slice1, y1
    #
    # @property
    # def ap_width_original(self):
    #     """ return original aperture widths """
    #     return np.median(self.ap_lower_interp - self.ap_upper_interp, axis=1)
    #
    # def ap_width_extended(self, n_extended=0):
    #     """ return extended aperture widths """
    #     return np.median(self.ap_lower_interp - self.ap_upper_interp, axis=1) \
    #            + 2 * n_extended
    #
    # def flatten(self, flat, method='median', **kwargs):
    #     """  to flatten FLAT
    #
    #     Parameters
    #     ----------
    #     flat:
    #         the image to be flattened
    #     method: string
    #         {"median"}
    #     kwargs:
    #         to be passed to specific flattening method
    #
    #     Returns
    #     -------
    #     flattened image
    #
    #     """
    #     # copy data
    #     im = copy(np.array(flat))
    #
    #     # flatten image
    #     im_flattened = self._flatten_median(im, **kwargs)
    #     im_sensitivity = im / im_flattened
    #     im_sensitivity = np.where(
    #         np.logical_or(np.logical_not(self.ap_region.sum(axis=0)),
    #                       np.logical_not(np.isfinite(im_sensitivity))),
    #         1.0, im_sensitivity)
    #
    #     return im_flattened, im_sensitivity

    # def _flatten_median(self, flat, kernel_size=(11, 3)):
    #     """ to flatten FLAT using medfilt2d
    #
    #     Parameters
    #     ----------
    #     flat
    #     kernel_size
    #
    #     Returns
    #     -------
    #
    #     """
    #     im_flattened = medfilt2d(flat, kernel_size=kernel_size)
    #     nht = np.fix(kernel_size[0] / 2)
    #     im_flattened[:, :nht] = flat[:, :nht]
    #     im_flattened[:, -nht:] = flat[:, -nht:]
    #     print(nht, im_flattened[:, -nht:])
    #     return im_flattened
    #
    # def extract_simple(self, im):
    #     """ simple extraction """
    #     return np.nansum((np.array([im]) * self.ap_weight), axis=1)
    #
    # def extract_median(self, im):
    #     """ simple extraction """
    #     data = np.array([im]) * self.ap_weight
    #     data = np.where(self.ap_weight > 0, data, np.nan)
    #     return np.nanmean(data, axis=1)
    #
    # def calculate_aperture_region(self, n_narrower=0):
    #     assert -9 < n_narrower < 9
    #
    #     x = np.arange(self.imshape[1], dtype=int)
    #     y = np.arange(self.imshape[0], dtype=int)
    #     mx, my = np.meshgrid(x, y)
    #     ap_region = np.zeros((self.n_ap, *self.imshape), dtype=bool)
    #     ap_weight = np.zeros((self.n_ap, *self.imshape), dtype=float)
    #     for i in range(self.n_ap):
    #         this_aperture_weight = np.zeros_like(mx, dtype=float)
    #         aperture_condition = np.logical_and(
    #             my > np.floor(self.ap_upper_interp[i]-n_narrower),
    #             my < np.ceil(self.ap_lower_interp[i]+n_narrower))
    #         min_disance_to_edge = np.min(np.abs(np.array(
    #             (my - self.ap_upper_interp[i] + n_narrower,
    #              self.ap_lower_interp[i] - my - n_narrower))), axis=0)
    #         this_aperture_weight = np.where(
    #             aperture_condition, min_disance_to_edge, this_aperture_weight)
    #         this_aperture_weight = np.where(
    #             this_aperture_weight > 1, 1.0, this_aperture_weight)
    #         ap_region[i] = aperture_condition
    #         ap_weight[i] = this_aperture_weight
    #
    #     return ap_region, ap_weight
    #
    # def update_aperture_region(self, n_narrower=0):
    #     self.ap_region, self.ap_weight = self.calculate_aperture_region(
    #         n_narrower=n_narrower)
    #     self.ap_width_original = np.median(
    #         self.ap_lower_interp - self.ap_upper_interp, axis=1) - 2*n_narrower

    # ####################################################################### #
    # Old decomposition code, decompose an order using an empirical spatial
    # profile and requires the position to be accurately determined, not useful
    # any more.
    # ####################################################################### #
    # def decomposition(self, im, im_var, vbinning=4, n_chunks=1,
    #                   smooth_kernel=3, cushion=4):
    #
    #     warnings.warn("DEPRECATED use stella instead ", DeprecationWarning)
    #
    #     im_var = np.abs(im_var)
    #     im_model = np.zeros_like(im, dtype=float)
    #     im_profile = np.zeros_like(im, dtype=float)
    #     blazes = np.zeros((self.n_ap, im.shape[1]), dtype=float)
    #
    #     # determine chunk start & stop
    #     chunk_len_typical = np.floor(im.shape[1]/n_chunks).astype(int)
    #     chunk_len = np.array([chunk_len_typical] * n_chunks).astype(int)
    #     chunk_len[-1] = im.shape[1] - np.sum(chunk_len[:-1]).astype(int)
    #     chunk_start = np.hstack((0, np.cumsum(chunk_len)[:-1])).astype(int)
    #     chunk_stop = np.cumsum(chunk_len).astype(int)
    #     print("@SONG: break each order to chunks with length: ", chunk_len)
    #
    #     for i_order in range(self.n_ap):
    #         # determine the section for this order
    #         ind_this_order = np.any(self.ap_region[i_order], axis=1)
    #
    #         # usually a cushion is needed in order to fit blaze function
    #         if cushion > 0:
    #             i_head, i_tail = np.where(ind_this_order)[0][[0, -1]]
    #             for i in range(cushion):
    #                 ind_this_order[np.max((0, i_head - 1 - i))] = True
    #                 ind_this_order[np.min((len(ind_this_order) - 1, i_tail + 1 + i))] = True
    #
    #         # section span in Y direction
    #         n_yspan = len(np.where(ind_this_order)[0])
    #         # aperture center offset relative to the center of this section
    #         this_order_center_ofst = - self.ap_center_interp[i_order] + \
    #                                  np.where(ind_this_order)[0][0] + 0.5 * n_yspan
    #         # cut the image to get the section for this order
    #         this_order = (im * self.ap_region[i_order])[ind_this_order]
    #         this_order_var = (im_var * self.ap_region[i_order])[ind_this_order]
    #         # this_order_x = self.mx[ind_this_order]
    #         # this_order_y = self.my[ind_this_order]
    #
    #         # break
    #         for i_chunk in range(n_chunks):
    #             this_order_chunk = this_order[:, chunk_start[i_chunk]:chunk_stop[i_chunk]]
    #             this_order_chunk_var = this_order_var[:, chunk_start[i_chunk]:chunk_stop[i_chunk]]
    #             profile_ofst = this_order_center_ofst[chunk_start[i_chunk]:chunk_stop[i_chunk]]
    #
    #             # decomposite this chunk
    #             this_order_chunk_profile_reconstructed, \
    #             this_order_chunk_reconstructed, blaze = \
    #                 decomposition_single_order_single_chunk(
    #                     this_order_chunk, this_order_chunk_var, profile_ofst,
    #                     vbinning=vbinning, smooth_kernel=smooth_kernel)
    #
    #             # co-add the chunk model to the final model
    #             im_model[ind_this_order, chunk_start[i_chunk]:chunk_stop[i_chunk]] \
    #                 += this_order_chunk_reconstructed
    #             # co-add the chunk profile to the final profile
    #             im_profile[ind_this_order, chunk_start[i_chunk]:chunk_stop[i_chunk]] \
    #                 += this_order_chunk_profile_reconstructed
    #             # collect blaze function
    #             blazes[i_order, chunk_start[i_chunk]:chunk_stop[i_chunk]] = blaze
    #         print("@SONG: modelling order ", i_order)
    #
    #     return im_model, im_profile, blazes


# def decomposition_single_order_single_chunk(this_order, this_order_var, profile_ofst, vbinning=4, smooth_kernel=None):
#     """ decompose an order (a chunk)
#
#     Parameters
#     ----------
#     this_order: (n_yspan, n_xall)
#         the image block
#     this_order_var: (n_yspan, n_xall)
#         the variance block
#     profile_ofst:
#         the offset relative to
#     vbinning: int
#         an oversampling factor
#     smooth_kernel: None | int
#         if not None, perform a smoothing on the blaze function
#         if None, do not smooth
#
#     Returns
#     -------
#
#     """
#     # make sure *this_order* is an array
#     this_order = np.array(this_order)
#
#     # determine profile coordinates, unbinned & binned
#     binning_step = 1.0 / vbinning
#     profile1d_coord_unbinned = np.arange(
#         0, this_order.shape[0], binning_step) - (vbinning - 1.) / (2. * vbinning)
#     profile1d_coord_unbinned_edge = np.hstack((
#         profile1d_coord_unbinned - 0.5 * binning_step,
#         profile1d_coord_unbinned[-1] + 0.5 * binning_step))
#     profile1d_coord_binned = np.arange(0, this_order.shape[0], 1.0)
#
#     # determine the profile using median
#     x = profile1d_coord_binned.reshape(-1, 1) + profile_ofst.reshape(1, -1)
#     blaze0 = np.sum(this_order, axis=0)
#     values = this_order / blaze0.reshape(1, -1)  # blaze0 only used here
#     # do *binned_statistic*
#     profile1d_unbinned, bin_edges, binnumber = binned_statistic(
#         x.flatten(), values.flatten(),
#         statistic=np.nanmedian,
#         bins=profile1d_coord_unbinned_edge)
#     profile1d_unbinned = np.where(np.isfinite(profile1d_unbinned),
#                                   profile1d_unbinned, 0.)
#     # force positivity
#     profile1d_unbinned = np.where(profile1d_unbinned > 0.,
#                                   profile1d_unbinned, 0.)
#     # force normalization
#     profile1d_unbinned /= np.sum(profile1d_unbinned)
#
#     # determine the blaze function (3 sigma rejection?)
#     this_order_profile_reconstructed = binning2dud(
#         offset_profile(profile1d_coord_unbinned, profile1d_unbinned,
#                        profile_ofst), vbinning)
#     this_order_profile_reconstructed /= \
#         np.sum(this_order_profile_reconstructed, axis=0).reshape(1, -1)
#
#     # determine the blaze function again
#     blaze1 = np.zeros((this_order.shape[1],), dtype=float)
#     for i_col in range(this_order.shape[1]):
#         blaze1[i_col], ier = leastsq(costfun, blaze0[i_col],
#                                      args=(this_order[:, i_col],
#                                            this_order_profile_reconstructed[:, i_col],
#                                            this_order_var[:, i_col]),
#                                      xtol=1.0
#                                      )
#     this_order_reconstructed = blaze1.reshape(1, -1) * this_order_profile_reconstructed
#
#     if smooth_kernel is None:
#         # no smoothing
#         return this_order_profile_reconstructed, this_order_reconstructed, blaze1
#     else:
#         # smooth blaze function and feed back to reconstructed image
#         lenend = np.int(np.ceil(smooth_kernel / 2.))
#         blaze2 = medfilt(blaze1, smooth_kernel)
#         blaze2[:lenend] = blaze1[:lenend]
#         blaze2[-lenend:] = blaze1[-lenend:]
#         this_order_reconstructed *= (blaze2 / blaze1).reshape(1, -1)
#         return this_order_profile_reconstructed, this_order_reconstructed, blaze2
#
#
# def costfun(x, profile_obs, profile_temp, profile_var):
#     weight = np.sqrt(profile_var)
#     # istd = np.where(np.isfinite(istd), istd, 0)
#     return (x * profile_temp - profile_obs) * weight
#
#
# def offset_profile(profile1d_coord, profile1d, ofsts):
#     """ reconstruct 2D profile with 1D profile and offsets """
#     f = interp1d(profile1d_coord, profile1d, kind='slinear', bounds_error=False, fill_value=0.)
#     profile2d_reconstructed = np.array([f(profile1d_coord + ofst) for ofst in ofsts])
#     return profile2d_reconstructed.T
#
#
# def binning2dlr(data, binning_factor=2):
#     """ rebin data at a given binning_factor (left-right) """
#     data = np.array(data)
#     binning_factor = np.int(binning_factor)
#     assert binning_factor >= 2
#     assert data.ndim == 2
#     assert np.mod(data.shape[1], binning_factor) == 0
#     binned_data = np.zeros((data.shape[0], np.int(data.shape[1] / binning_factor)), dtype=data.dtype)
#     for i in range(binning_factor):
#         binned_data += data[:, i::binning_factor]
#     return binned_data
#
#
# def binning2dud(data, binning_factor=2):
#     """ rebin data at a given binning_factor (up-down) """
#     data = np.array(data)
#     binning_factor = np.int(binning_factor)
#     assert binning_factor >= 2
#     assert data.ndim == 2
#     assert np.mod(data.shape[0], binning_factor) == 0
#     binned_data = np.zeros((np.int(data.shape[0] / binning_factor), data.shape[1]), dtype=data.dtype)
#     for i in range(binning_factor):
#         binned_data += data[i::binning_factor, :]
#     return binned_data
#
#
# class Scaler(object):
#
#     def __init__(self, arr, range_=(-1, 1)):
#         """ to scale arr to a specific range """
#         arr = np.array(arr)
#         range_ = np.array(range_)
#
#         scale = np.abs((range_[0] - range_[-1]) / (arr[-1] - arr[0]))
#         arr_scale = arr * scale
#         ofst = np.min(arr_scale) - np.min(range_)
#         arr_scale_ofst = arr_scale - ofst
#
#         self.arr = arr
#         self.arr_scaled_ofst = arr_scale_ofst
#         self.scale_ = scale
#         self.ofst_ = ofst
#
#     def scale(self, data):
#         return np.array(data) * self.scale_ - self.ofst_
#
#     def scale_back(self, data):
#         return (np.array(data) + self.ofst_) / self.scale_


def test_aperture():
    from twodspec.aperture import Aperture
    from astropy.io import fits
    import matplotlib.pyplot as plt
    plt.rcParams.update({"font.size": 15})
    flat = fits.getdata("/Users/cham/projects/song/star_spec/20191105/night/ext/masterflat_20191105_slit5.fits")

    ap = Aperture.trace(flat, method="naive", polydeg=4, sigma=7, maxdev=7, ap_width=15)
    ap = Aperture.trace(flat, method="canny", polydeg=4, sigma=7, maxdev=7, ap_width=15)
    return ap

    # from twodspec import extract
    # _blz, _norm = extract.make_normflat(flat, ap, ap_width=12)
    #
    # figure();    imshow(_norm)
    #
    # rextr = extract.extract_all(flat, ap, ap_width=12)
    #
    # # %%
    # starbg = ap.background(star, q=(5, 5))
    # rextr = extract.extract_all(star - starbg, ap, ap_width=10)
    #
    # figure();
    # plot(star[1200]);
    # plot(starbg[1200])
    # # %%
    # figure();
    # plot(np.arange(2048), rextr["spec_sum"][10].T / _blz[10].T)
    # plot(np.arange(2048) + 2048, rextr["spec_sum"][11].T / _blz[11].T)
    #
    # # figure();plot(_blz.T)
    # # %%
    # star0 = fits.getdata("/media/cham/Ubuntu 17.1/obj4blue_multi.fits")
    # # %%
    #
    # flux0 = star0[1, 5, :].T
    # flux1 = rextr["spec_extr"][-5].T
    # flux2 = rextr["spec_extr2"][-5].T
    # wave = np.arange(2048)
    #
    # flux0n, _ = normalize_spectrum_iter(wave, flux0, niter=3)
    # flux1n, _ = normalize_spectrum_iter(wave, flux1, niter=3)
    # flux2n, _ = normalize_spectrum_iter(wave, flux2, niter=3)


if __name__ == "__main__":
    test_aperture()



# -*- coding: utf-8 -*-
"""

Author
------
Bo Zhang

Email
-----
bozhang@nao.cas.cn

Created on
----------
- Fri Apr 28 14:00:00 2017

Modifications
-------------
-

Aims
----
- ways to trace apertures

"""

import os

import numpy as np
from astropy.io import fits
from matplotlib import pyplot as plt
from scipy import signal
# from skimage import feature


# not used ??
# def trace_canny_row(flat, details=False, *args, **kwargs):
#     """ trace apertures using canny edge method in skimage for row apertures
#
#     Parameters
#     ----------
#     flat: numpy.ndarray
#         the image used to trace apertures, usually FLAT is used
#     details: bool
#         if True, return details
#     args, kwargs:
#         args & kwargs passed to skimage.feature.canny()
#
#     """
#
#     # 0. conservatively, transform to numpy.array
#     flatdata = np.array(flat)
#     n_row, n_col = flatdata.shape
#
#     # 1. find the edges using skimage.feature.canny()
#     # edge = feature.canny(flat.data, sigma=3.0, low_threshold=.05,
#     #                      use_quantiles=True)
#     edge = feature.canny(flatdata, *args, **kwargs)
#
#     # 2. pick out reasonable edges as aperture edges
#     # criterion: from column 1 to column -1
#     edge_good = []
#     for i_row in range(n_row):
#         i_col = 1
#         if edge[i_row, i_col]:
#             ap_list = []
#             ofst_row = 0
#             ap_list.append((i_row + ofst_row))
#             for i_col in range(2, n_col - 1):
#                 if edge[i_row + ofst_row, i_col]:
#                     ap_list.append((i_row + ofst_row))
#                 elif i_row + ofst_row + 1 < n_row - 1 and edge[
#                                     i_row + ofst_row + 1, i_col]:
#                     ofst_row += 1
#                     ap_list.append((i_row + ofst_row))
#                 elif i_row + ofst_row - 1 > 0 and edge[
#                                     i_row + ofst_row - 1, i_col]:
#                     ofst_row -= 1
#                     ap_list.append((i_row + ofst_row))
#                 else:
#                     # ap_list = []
#                     # ofst_row = 0
#                     break
#         if i_col >= flat.data.shape[1] - 2:
#             edge_good.append(ap_list)
#
#     ap_col = np.arange(1, n_col - 1, dtype=int)
#     ap_row = np.array(edge_good, dtype=int)
#
#     # 3. classify edges to {upper, lower} category
#     isupper = np.array(
#         [np.percentile(flat[ap_row_ - 3, ap_col], 50) < np.percentile(
#             flat[ap_row_ + 3, ap_col], 50) for ap_row_ in ap_row])
#
#     # 4. adopt good edges as true apertures
#     isadopted = np.zeros_like(isupper, dtype=bool)
#
#     # find an approporiate start for aperture
#     iap_start = 0
#     ap_upper = []
#     ap_lower = []
#     while True:
#         # if out of range, break
#         if iap_start >= len(isupper) - 1:
#             break
#
#         # grab apertures from the left
#         if isupper[iap_start] and not isupper[iap_start + 1]:
#             # and 10<np.median(ap_row[iap_start+1]-ap_row[iap_start])<30
#             # then, this is a good aperture
#             ap_upper.append(ap_row[iap_start])
#             ap_lower.append(ap_row[iap_start + 1])
#             isadopted[iap_start:iap_start + 2] = True
#             iap_start += 2
#         else:
#             # then, this is not a good aperture
#             print("@trace_canny: bad aperture for iap_start = ", iap_start)
#             iap_start += 1
#
#     # transform to numpy.array format
#     ap_upper = np.array(ap_upper)
#     ap_lower = np.array(ap_lower)
#     # fill left & right ends
#     ap_lower = np.hstack((ap_lower[:, 0].reshape(-1, 1), ap_lower,
#                           ap_lower[:, -1].reshape(-1, 1)))
#     ap_upper = np.hstack((ap_upper[:, 0].reshape(-1, 1), ap_upper,
#                           ap_upper[:, -1].reshape(-1, 1)))
#     # number of apertures
#     n_ap = ap_upper.shape[0]
#
#     # 5. assign results to Aperture object
#     results = dict(
#         # number of apertures found
#         n_ap=n_ap,
#
#         # edge int arrays, upper & lower seperated
#         ap_lower=ap_lower,
#         ap_upper=ap_upper,
#         ap_center=(ap_lower + ap_upper)/2
#     )
#
#     if not details:
#         return results
#     else:
#         details = dict(
#             # raw edges
#             edge=edge,
#
#             # edge int arrays, upper & lower together
#             ap_col=ap_col,
#             ap_row=ap_row,
#
#             # adopted edges & info
#             edge_good=edge_good,
#             isupper=isupper,
#             isadopted=isadopted,
#         )
#         return results, details
#
#
# def trace_canny_col(flat, details=False, verbose=True, *args, **kwargs):
#     """ trace apertures using canny method in skimage for column apertures
#
#     Parameters
#     ----------
#     flat: numpy.ndarray
#         the image used to trace apertures, usually FLAT is used
#     details: bool
#         if True, return details
#     args, kwargs:
#         args & kwargs passed to skimage.feature.canny()
#
#     """
#
#     # 0. conservatively, transform to numpy.array
#     flatdata = np.array(flat)
#     n_row, n_col = flatdata.shape
#
#     # 1. find the edges using skimage.feature.canny()
#     # edge = feature.canny(flat.data, sigma=3.0, low_threshold=.05,
#     #                      use_quantiles=True)
#     edge = feature.canny(flatdata, *args, **kwargs)
#
#     # 2. pick out reasonable edges as aperture edges
#     # criterion: from column 1 to column -1
#     n_row, n_col = np.rot90(flat.data).shape
#     edge_good = []
#     for i_col in range(n_row):
#         i_row = 1
#         if edge[i_row, i_col]:
#             ap_list = []
#             ofst_col = 0
#             ap_list.append(i_col + ofst_col)
#             for i_row in range(2, n_col - 1):
#                 # starts from the first row
#                 if edge[i_row, i_col + ofst_col]:
#                     # if the same column in the next row is True
#                     ap_list.append(i_col + ofst_col)
#                 elif i_col + ofst_col + 1 <= n_col - 1 \
#                         and edge[i_row, i_col + ofst_col + 1]:
#                     # offset = +1
#                     ofst_col += 1
#                     ap_list.append(i_col + ofst_col)
#                 elif i_col + ofst_col - 1 >= 0 \
#                         and edge[i_row, i_col + ofst_col - 1]:
#                     # offset = -1
#                     ofst_col -= 1
#                     ap_list.append(i_col + ofst_col)
#                 else:
#                     break
#         if i_row >= flat.data.shape[1] - 2:
#             edge_good.append(ap_list)
#
#     ap_row = np.arange(1, n_row - 1, dtype=int)
#     ap_col = np.array(edge_good, dtype=int)
#
#     # 3. classify edges to {upper, lower} category
#     isupper = np.array(
#         [np.percentile(flat[ap_row, ap_col_ - 3], 50) < np.percentile(
#             flat[ap_row, ap_col_ + 3], 50) for ap_col_ in ap_col])
#
#     # 4. adopt good edges as true apertures
#     isadopted = np.zeros_like(isupper, dtype=bool)
#
#     # find an approporiate start for aperture
#     iap_start = 0
#     ap_upper = []
#     ap_lower = []
#     while True:
#         # if out of range, break
#         if iap_start >= len(isupper) - 1:
#             break
#
#         # grab apertures from the left
#         if isupper[iap_start] and not isupper[iap_start + 1]:
#             # and 10<np.median(ap_row[iap_start+1]-ap_row[iap_start])<30
#             # then, this is a good aperture
#             ap_upper.append(ap_col[iap_start])
#             ap_lower.append(ap_col[iap_start + 1])
#             isadopted[iap_start:iap_start + 2] = True
#             iap_start += 2
#         else:
#             # then, this is not a good aperture
#             if verbose:
#                 print("@trace_canny: bad aperture for iap_start = ", iap_start)
#             iap_start += 1
#
#     # transform to numpy.array format
#     ap_upper = np.array(ap_upper)
#     ap_lower = np.array(ap_lower)
#     # fill left & right ends
#     ap_lower = np.hstack((ap_lower[:, 0].reshape(-1, 1), ap_lower,
#                           ap_lower[:, -1].reshape(-1, 1)))
#     ap_upper = np.hstack((ap_upper[:, 0].reshape(-1, 1), ap_upper,
#                           ap_upper[:, -1].reshape(-1, 1)))
#     # number of apertures
#     n_ap = ap_upper.shape[0]
#
#     # 5. assign results to Aperture object
#     results = dict(
#         # number of apertures found
#         n_ap=n_ap,
#
#         # edge int arrays, upper & lower seperated
#         ap_lower=ap_lower,
#         ap_upper=ap_upper,
#         ap_center=(ap_lower + ap_upper) / 2
#     )
#
#     if not details:
#         return results
#     else:
#         details = dict(
#             # raw edges
#             edge=edge,
#
#             # edge int arrays, upper & lower together
#             ap_col=ap_col,
#             ap_row=ap_row,
#
#             # adopted edges & info
#             edge_good=np.array(edge_good),
#             isupper=isupper,
#             isadopted=isadopted,
#         )
#         return results, details


def gaussian_kernel(sigma=2., n_sigma=5):
    """ generate Gaussian kernel

    Parameters
    ----------
    sigma : float, optional
        The sigma of the Gaussian. The default is 2..
    n_sigma : flaot, optional
        The width of the Gaussian in terms of sigma. The default is 5.

    Returns
    -------
    y : ndarray
        The Gaussian kernel array.

    """
    npix = np.int(n_sigma * sigma) + 1
    x = np.arange(-npix, npix + 1, 1)
    y = np.exp(-0.5 * (x / sigma) ** 2)
    y /= np.sum(y)
    return y


# not used
def box_kernel(fullwidth=10):
    npix = np.int(fullwidth)
    kernel = np.ones(npix)
    kernel /= np.sum(kernel)
    return kernel


def trace_naive_max(flat, sigma=7., maxdev=0, irow_start=None):
    """ trace aperture using naive method -- along local max

    Parameters
    ----------
    flat : ndarray
        FLAT.
    sigma : float, optional
        The sigma of Gaussian kernel. The default is 7.
    maxdev : float, optional
        The max deviation of local max while tracing. The default is 0.
    irow_start :
        The number of the starting row

    Returns
    -------
    ap_center : ndarray (n_ap, n_pix)
        The array for center of aperture.

    """
    flat = np.asarray(flat, float)
    n_row, n_col = flat.shape
    if irow_start is None:
        i_row_center = np.int(n_row / 2)
    else:
        i_row_center = irow_start

    # smooth flat along spacial axis
    flat_smooth = np.zeros_like(flat)
    gkernel = gaussian_kernel(sigma)
    for i_row in range(n_row):
        flat_smooth[i_row] = signal.fftconvolve(flat[i_row], gkernel, mode="same")

    # find local max
    flat_smooth_diff = np.diff(flat_smooth, axis=1)
    flat_localmax = np.zeros_like(flat, bool)
    flat_localmax[:, 1:-1] = (flat_smooth_diff[:, 1:] < 0) & (flat_smooth_diff[:, :-1] > 0)

    # start from center row
    ind_col_localmax, = np.where(flat_localmax[i_row_center])
    n_ap_try = len(ind_col_localmax)
    ap_center_try = np.ones((n_ap_try, n_row)) * np.nan
    ap_center_try[:, i_row_center] = ind_col_localmax

    # find local max col each row
    each_row_localmax = [np.where(flat_localmax[i_row])[0] for i_row in range(n_row)]

    # loop for trials
    if maxdev <= 0 or maxdev is None:
        maxdev = sigma
    for i_ap in range(n_ap_try):
        for i_row in range(i_row_center - 1, 0 - 1, -1):
            this_d = np.abs(each_row_localmax[i_row] - ap_center_try[i_ap, i_row + 1])
            if np.min(this_d) <= maxdev:
                ap_center_try[i_ap, i_row] = each_row_localmax[i_row][np.argmin(this_d)]
            else:
                break
        for i_row in range(i_row_center + 1, n_row, 1):
            this_d = np.abs(each_row_localmax[i_row] - ap_center_try[i_ap, i_row - 1])
            if np.min(this_d) <= maxdev:
                ap_center_try[i_ap, i_row] = each_row_localmax[i_row][np.argmin(this_d)]
            else:
                break
    ind_good_ap = np.sum(np.isfinite(ap_center_try), axis=1) == n_row
    n_ap = np.sum(ind_good_ap)
    ap_center = ap_center_try[ind_good_ap]

    return ap_center


def trace_naive_min(flat, sigma=7., maxdev=0):
    """ trace aperture using naive method -- along local min

    Parameters
    ----------
    flat : ndarray
        FLAT.
    sigma : float, optional
        The sigma of Gaussian kernel. The default is 7.
    maxdev : float, optional
        The max deviation of local max while tracing. The default is 0.

    Returns
    -------
    ap_center : ndarray (n_ap, n_pix)
        The array for center of aperture..

    """
    flat = np.asarray(flat, float)
    n_row, n_col = flat.shape
    i_row_center = np.int(n_row / 2)

    # smooth flat
    flat_smooth = np.zeros_like(flat)
    gkernel = gaussian_kernel(sigma)
    for i_row in range(n_row):
        flat_smooth[i_row] = signal.fftconvolve(flat[i_row], gkernel, mode="same")

    # find local max
    flat_smooth_diff = np.diff(flat_smooth, axis=1)
    flat_localmax = np.zeros_like(flat, bool)
    flat_localmax[:, 1:-1] = (flat_smooth_diff[:, 1:] > 0) & (flat_smooth_diff[:, :-1] < 0)

    # start from center row
    ind_col_localmax, = np.where(flat_localmax[i_row_center])
    n_ap_try = len(ind_col_localmax)
    ap_center_try = np.ones((n_ap_try, n_row)) * np.nan
    ap_center_try[:, i_row_center] = ind_col_localmax

    # each row local max col
    each_row_localmax = [np.where(flat_localmax[i_row])[0] for i_row in range(n_row)]

    # loop for trials
    if maxdev <= 0 or maxdev is None:
        maxdev = sigma
    for i_ap in range(n_ap_try):
        for i_row in range(i_row_center - 1, 0 - 1, -1):
            this_d = np.abs(each_row_localmax[i_row] - ap_center_try[i_ap, i_row + 1])
            if np.min(this_d) <= maxdev:
                ap_center_try[i_ap, i_row] = each_row_localmax[i_row][np.argmin(this_d)]
            else:
                break
        for i_row in range(i_row_center + 1, n_row, 1):
            this_d = np.abs(each_row_localmax[i_row] - ap_center_try[i_ap, i_row - 1])
            if np.min(this_d) <= maxdev:
                ap_center_try[i_ap, i_row] = each_row_localmax[i_row][np.argmin(this_d)]
            else:
                break
    ind_good_ap = np.sum(np.isfinite(ap_center_try), axis=1) == n_row
    # n_ap = np.sum(ind_good_ap)
    ap_center = ap_center_try[ind_good_ap]

    return ap_center


def test_trace_naive_max():
    plt.rcParams.update({"font.size": 15})
    flat = fits.getdata("/Users/cham/projects/song/star_spec/20191105/night/ext/masterflat_20191105_slit5.fits")
    ap_center = trace_naive_max(flat, sigma=7, maxdev=10)
    fig = plt.figure(figsize=(8,6))
    plt.plot(ap_center.T, np.arange(2048), "-", c="w",label="apertures",lw=1)
    plt.imshow(np.log10(flat),cmap=plt.cm.jet)
    # plt.set_xticks(np.linspace(0,2047,))
    plt.colorbar()
    plt.xlabel("CCD X coordinate")
    plt.ylabel("CCD Y coordinate")
    fig.tight_layout()
    # fig.savefig(os.getenv("HOME")+"/PycharmProjects/songcn/figs/test_trace_naive_max.pdf")
    print(ap_center.shape)
    return


if __name__ == "__main__":
    test_trace_naive_max()


# -*- coding: utf-8 -*-
"""

Author
------
Bo Zhang

Email
-----
bozhang@nao.cas.cn

Created on
----------
- Fri Nov 25 12:53:24 2016

Modifications
-------------
-

Aims
----
- utils for apertures

"""

from copy import copy

import numpy as np
from joblib import Parallel, delayed
from scipy.interpolate import interp1d
from scipy.signal import medfilt2d
from skimage.filters import gaussian

from . import ccdproc_mod as ccdproc
from .normalization import normalize_spectra_block


# ################################# #
#      find starting points
# ################################# #

def find_local_maximas(pixels, n_adj=2, n_smooth=1, n_sep=1):
    """ find all local maximas

    Parameters
    ----------
    pixels:
        a slice of CCD pixels
    n_adj:
        number of adjacent pixels for maximas
    n_smooth:
        running mean smooth window width
    n_sep:
        the separation since last maxima

    Returns
    -------
    ind of maximas

    """
    pixels = np.array(pixels).flatten()
    n_pix = len(pixels)

    ind_max = np.zeros_like(pixels, dtype=bool)
    # smooth
    for i in np.arange(n_smooth, n_pix - n_smooth):
        pixels[i] = np.mean(np.array(pixels[i - n_smooth:i + 1 + n_smooth]))
    # find maximas
    for i in np.arange(n_adj, n_pix - n_adj):
        ind_max[i] = np.argmax(
            np.array(pixels[i - n_adj:i + 1 + n_adj])) == n_adj and np.sum(
            ind_max[i - n_sep:i]) < 1

    return ind_max


def find_local_minimas(pixels, n_adj=2, n_smooth=1, n_sep=10):
    """ find all local minimas

    Parameters
    ----------
    pixels:
        a slice of CCD pixels
    n_adj:
        number of adjacent pixels for minimas
    n_smooth:
        running mean smooth window width
    n_sep:
        the separation since last minima

    Returns
    -------
    ind of minimas

    """
    pixels = np.array(pixels).flatten()
    n_pix = len(pixels)

    ind_min = np.zeros_like(pixels, dtype=bool)
    # smooth
    for i in np.arange(n_smooth, n_pix - n_smooth):
        pixels[i] = np.mean(np.array(pixels[i - n_smooth:i + 1 + n_smooth]))
    # find minimas
    for i in np.arange(n_adj, n_pix - n_adj):
        ind_min[i] = np.argmin(
            np.array(pixels[i - n_adj:i + 1 + n_adj])) == n_adj and np.sum(
            ind_min[i - n_sep:i]) < 1

    return ind_min


def find_mmax_mmin(im, start_col=2100, n_adj=7, n_smooth=3, n_sep=5,
                   verbose=False):
    """ find reasonable maximas and minimas (cross-kick)

    Parameters
    ----------
    im
    start_col

    Returns
    -------

    """
    start_col_slice = np.sum(im[:, start_col][:, None], axis=1)

    imax = find_local_maximas(start_col_slice, n_adj=n_adj, n_smooth=n_smooth,
                              n_sep=n_sep)
    imin = find_local_minimas(start_col_slice, n_adj=n_adj, n_smooth=n_smooth,
                              n_sep=n_sep)

    # no pixel could be both max&min
    try:
        assert np.sum(imax * imin) == 0
    except:
        raise (ValueError("@TWODSPEC: imax*imin != 0"))

    # cross kick
    smax = np.where(imax)[0]
    smin = np.where(imin)[0]
    smmax = []
    smmin = []
    for i in range(len(smax) - 1):
        imin_this = imin[smax[i]:smax[i + 1]]  # ind array for local mins
        smin_this = np.where(imin_this)[0]  # sub array for local mins
        if len(smin_this) == 1:
            smmin.append(smin_this[0] + smax[i])
        elif len(smin_this) > 1:
            # more than 1 local minimas, find the minimal minima
            smmin.append(smin_this[np.argmin(
                start_col_slice[smax[i]:smax[i + 1]][smin_this])] + smax[i])
    # smin = smmin
    for i in range(len(smmin) - 1):
        imax_this = imax[smmin[i]:smmin[i + 1]]  # ind array for local maxs
        smax_this = np.where(imax_this)[0]  # sub array for local maxs
        if len(smax_this) == 1:
            smmax.append(smax_this[0] + smmin[i])
        elif len(smax_this) > 1:
            # more than 1 local maximas, find the maximal maxima
            smmax.append(smax_this[np.argmax(
                start_col_slice[smmin[i]:smmin[i + 1]][smax_this])] + smmin[i])

    if verbose:
        print("@TWODSPEC: %s possible apertures found!" % len(smmax))

    return smmax, smmin


# ################################# #
#      find apertures
# ################################# #

def find_apertures(im, start_col=2100, max_drift=5, max_apwidth=10,
                   n_pix_goodap=1500, n_adj=7, n_smooth=3, n_sep=5, c=3,
                   verbose=False):
    """ find apertures from image

    Parameters
    ----------
    im:
        image
    start_col:
        start column
    max_drift:
        max_drift in finding an aperture
    max_apwidth:
        local comparison width
    n_pix_goodap:
        a good aperture should be more than this number of pixels

    Returns
    -------
    ymmax_goodap: ndarray
        the y-pixel values for good apertures
    """
    # gaussian smooth if c>0
    if c > 0:
        im = ccdproc.CCDData(gaussian(im, sigma=(c, 0)), unit=im.unit)

    # find max & min
    smmax, smmin = find_mmax_mmin(im, start_col=start_col, n_adj=n_adj,
                                  n_smooth=n_smooth, n_sep=n_sep)

    # initialize results
    ymmax = np.zeros((len(smmax), im.shape[1]))
    ymmax[:, start_col] = smmax
    ymmin = np.zeros((len(smmin), im.shape[1]))
    ymmin[:, start_col] = smmin

    # tracing apertures
    for i_ap in range(ymmax.shape[0]):
        for i_col in np.arange(start_col + 1, im.shape[1]):
            y0 = ymmax[i_ap, i_col - 1]
            # print(np.int(np.max((0, y0 - max_apwidth))))
            # print(np.int(np.min((im.shape[0], y0 + 1 + max_apwidth))))
            # print(np.int(i_col))
            # print(np.int(np.max((0, y0 - max_apwidth))))
            # print(np.int(np.min((im.shape[0], y0 + 1 + max_apwidth))))
            # print(np.int(i_col - 1))
            # print(y0)
            # print(max_apwidth)
            # print("PROD", im[np.int(np.max((0, y0 - max_apwidth))):np.int(np.min(
            #     (im.shape[0], y0 + 1 + max_apwidth))), np.int(i_col)].data *
            #                im[np.int(np.max((0, y0 - max_apwidth))):np.int(np.min(
            #                    (im.shape[0], y0 + 1 + max_apwidth))),
            #                np.int(i_col - 1)].data + y0 - max_apwidth)
            y1 = np.argmax(im[np.int(np.max((0, y0 - max_apwidth))):np.int(np.min(
                (im.shape[0], y0 + 1 + max_apwidth))), np.int(i_col)].data *
                           im[np.int(np.max((0, y0 - max_apwidth))):np.int(np.min(
                               (im.shape[0], y0 + 1 + max_apwidth))),
                           np.int(i_col - 1)].data) + y0 - max_apwidth
            if np.abs(y1 - y0) < max_drift:
                # good ap, continue
                ymmax[i_ap, i_col] = y1
            else:
                break
        for i_col in np.arange(start_col - 1, 0, -1):
            y0 = ymmax[i_ap, i_col + 1]
            y1 = np.argmax(im[np.max((0, y0 - max_apwidth)):np.min(
                (im.shape[0], y0 + 1 + max_apwidth)), i_col].data *
                           im[np.max((0, y0 - max_apwidth)):np.min(
                               (im.shape[0], y0 + 1 + max_apwidth)),
                           i_col + 1].data) + y0 - max_apwidth
            if np.abs(y1 - y0) < max_drift:
                # good ap, continue
                ymmax[i_ap, i_col] = y1
            else:
                break
        if verbose:
            print("@TWODSPEC: tracing aperture [%s] " % i_ap)
    #
    # for i_ap in range(ymmin.shape[0]):
    #     for i_col in np.arange(start_col + 1, im.shape[1]):
    #         y0 = ymmin[i_ap, i_col - 1]
    #         y1 = np.argmin(im[np.max((0, y0 - max_apwidth)):np.min(
    #             (im.shape[0], y0 + 1 + max_apwidth)), i_col].data *
    #                        im[np.max((0, y0 - max_apwidth)):np.min(
    #                            (im.shape[0], y0 + 1 + max_apwidth)),
    #                        i_col - 1].data) + y0 - max_apwidth
    #         if np.abs(y1 - y0) < max_drift:
    #             # good ap, continue
    #             ymmin[i_ap, i_col] = y1
    #         else:
    #             break
    #     for i_col in np.arange(start_col - 1, 0, -1):
    #         y0 = ymmin[i_ap, i_col + 1]
    #         y1 = np.argmin(im[np.max((0, y0 - max_apwidth)):np.min(
    #             (im.shape[0], y0 + 1 + max_apwidth)), i_col].data *
    #                        im[np.max((0, y0 - max_apwidth)):np.min(
    #                            (im.shape[0], y0 + 1 + max_apwidth)),
    #                        i_col + 1].data) + y0 - max_apwidth
    #         if np.abs(y1 - y0) < max_drift:
    #             # good ap, continue
    #             ymmin[i_ap, i_col] = y1
    #         else:
    #             break
    #     print i_ap

    ind_goodap = np.sum(ymmax > 0, axis=1) > n_pix_goodap
    ymmax_goodap = ymmax[ind_goodap, :]
    print("@TWODSPEC: number of good aps (max) = %s " % np.sum(ind_goodap))

    #    ind_goodap = np.sum(ymmax>0, axis=1)>1000
    #    ymmin_goodap = ymmin[ind_goodap, :]
    #    print("@TWODSPEC: gumber of good aps (min)", np.sum(ind_goodap))

    return ymmax_goodap  # , ymmin_goodap


# ################################# #
#      combine & group apertures
# ################################# #

def combine_apertures(imlist, n_jobs=10, find_aps_param_dict=None,
                      verbose=False):
    """ combine apertures found from different FLAT images

    Parameters
    ----------
    imlist: list
        list of FLAT
    find_aps_param_dict:
        the Parameters used in *find_apertures*
    n_jobs:
        n_jobs for finding apertures in parallel

    Returns
    -------
    ap_combine:
        the y-coordinates of good apertures found from FLAT list

    """
    if find_aps_param_dict is None:
        find_aps_param_dict = dict(start_col=2100, max_drift=9, max_apwidth=13,
                                   n_pix_goodap=1000, n_adj=7, n_smooth=3,
                                   n_sep=5, c=3)
    ap_list = Parallel(n_jobs=n_jobs, verbose=2)(
        delayed(find_apertures)(im, verbose=verbose, **find_aps_param_dict)
        for im in imlist)

    if verbose:
        print("@TWODSPEC: the numbers of apertures found are: ",
              [ap_.shape[0] for ap_ in ap_list])

    ap_combine = np.vstack(ap_list)
    return ap_combine


def group_apertures(ap_comb, start_col=2100, order_dist=10, verbose=False):
    """ group combined apertures to unique apertures

    Parameters
    ----------
    ap_comb:
        combined apertures
    start_col:
        the starting column number
    order_dist:
        the typical distance between orders

    Returns
    -------
    cheb_coefs: ndarray
        the chebyshev polynomial coefs
    ap_uorder_interp: ndarray
        the y-coordinates of unique orders

    """
    # extract shape
    naps = ap_comb.shape[0]
    npix_disp = ap_comb.shape[1]

    # initialize
    x = np.arange(npix_disp)
    yi_start_col = np.zeros(naps)

    # rough chebyshev fit & interpolate
    for i in range(naps):
        y = ap_comb[i]
        c = np.polynomial.chebyshev.chebfit(x, y, 2, w=y > 0)
        yi_start_col[i] = np.polynomial.chebyshev.chebval(x[start_col], c)
    yi_start_col_sorted = np.sort(yi_start_col)
    yi_start_col_sortarg = np.argsort(yi_start_col)

    # find unique orders
    order_arg_list = []
    c = 0
    for i_uorder in range(naps):
        this_order_yi = yi_start_col_sorted[c]
        this_order_ind = np.abs(
            yi_start_col_sorted - this_order_yi) < order_dist
        this_order_arg = yi_start_col_sortarg[this_order_ind]
        order_arg_list.append(this_order_arg)
        c += len(this_order_arg)
        if c >= len(yi_start_col_sorted):
            break

    # fit & extrapolate
    n_uorder = len(order_arg_list)
    cheb_coefs = []
    ap_uorder_interp = np.zeros((n_uorder, npix_disp))
    for i_uorder in range(n_uorder):
        n_suborder = len(order_arg_list[i_uorder])
        x_data = np.repeat(x[:, None].T, n_suborder, axis=0).flatten()
        y_data = ap_comb[order_arg_list[i_uorder], :].flatten()
        c = np.polynomial.chebyshev.chebfit(x_data, y_data, 2, w=y_data > 0)
        cheb_coefs.append(c)
        ap_uorder_interp[i_uorder] = np.polynomial.chebyshev.chebval(x, c)

    if verbose:
        print("@TWODSPEC: %s unique orders found!" % n_uorder)

    return cheb_coefs, ap_uorder_interp


# ################################# #
#      extract 1d spectra
# ################################# #

def extract_1dspec(im, ap_uorder_interp, ap_width=7, verbose=False):
    naps = ap_uorder_interp.shape[0]
    npix = ap_uorder_interp.shape[1]
    ap_int = ap_uorder_interp.astype(int)

    spec = np.zeros((naps, npix), dtype=float)
    sat_mask = np.zeros((naps, npix), dtype=bool)

    for i in range(naps):
        for j in range(npix):
            yc = ap_int[i, j]
            data = im[yc - ap_width:yc + ap_width + 1, j]
            spec[i, j] = np.sum(data)
            sat_mask[i, j] = np.any(data > 60000)

        if verbose:
            print("@TWODSPEC: extracting 1d spec for order [{0}]".format(i))

    return spec, sat_mask


# ################################# #
#      find ind for all orders
# ################################# #

def find_ind_order(ap_uorder_interp, ccd_shape, edge_len=(10, 20)):
    # find bounds for apertures
    ap_uorder_bounds = np.vstack(
        (ap_uorder_interp[0] - edge_len[0],
         np.diff(ap_uorder_interp, axis=0) * .5 + ap_uorder_interp[:-1],
         ap_uorder_interp[-1] + edge_len[1]))

    # initialize ind_order
    ind_order = np.zeros(ccd_shape, dtype=int)

    # generate coordinates
    x, y = np.arange(ccd_shape[1]), np.arange(ccd_shape[0])
    mx, my = np.meshgrid(x, y)

    # for each apertures
    for i_order in range(ap_uorder_bounds.shape[0] - 1):
        this_order_bound_l = ap_uorder_bounds[i_order, :].reshape(1, -1)
        this_order_bound_u = ap_uorder_bounds[i_order + 1].reshape(1, -1)
        ind_this_order = np.logical_and(my > this_order_bound_l,
                                        my < this_order_bound_u)
        ind_order[ind_this_order] = i_order
        print ("@TWODSPEC: marking pixels of order [%s]" % i_order)
    # for edges
    this_order_bound_l = -1
    this_order_bound_u = ap_uorder_bounds[0].reshape(1, -1)
    ind_this_order = np.logical_and(my > this_order_bound_l,
                                    my < this_order_bound_u)
    ind_order[ind_this_order] = -1

    this_order_bound_l = ap_uorder_bounds[-1].reshape(1, -1)
    this_order_bound_u = ccd_shape[0]
    ind_this_order = np.logical_and(my > this_order_bound_l,
                                    my < this_order_bound_u)
    ind_order[ind_this_order] = -2

    return ind_order


# ############################################### #
#      combine flat according to max counts
# ############################################### #

def combine_flat(flat_list, ap_uorder_interp, sat_count=50000, p=95):
    """ combine flat according to max value under sat_count

    Parameters
    ----------
    flat_list
    ap_uorder_interp
    sat_count
    p

    Returns
    -------

    """
    # find ind_order
    ind_order = find_ind_order(ap_uorder_interp, flat_list[0].shape)

    # unique orders
    uorder = np.unique(ind_order)
    uorder_valid = uorder[uorder >= 0]

    # find max for each order
    im_max = np.vstack(
        [find_each_order_max(im_, ind_order, p) for im_ in flat_list])

    # combine flat
    flat_final = np.zeros_like(flat_list[0])
    flat_origin = np.zeros_like(flat_list[0], dtype=int)
    for uorder_ in uorder:
        # find pixels of this uorder
        ind_this_order = ind_order == uorder_
        # which one should be used for this uorder
        i_im = find_max_under_saturation(im_max[:, uorder_],
                                         sat_count=sat_count)
        # fill the findal image with data from that image
        flat_final = np.where(ind_this_order, flat_list[i_im], flat_final)
        flat_origin[ind_this_order] = i_im
        print("@TWODSPEC: filling data: image [%s] --> uorder [%s]" %
              (i_im, uorder_))

    return flat_final, flat_origin


def find_max_under_saturation(max_vals, sat_count=45000):
    """ find the image with maximum value under saturation count number

    Parameters
    ----------
    max_vals
    sat_count

    Returns
    -------
    the ID of max value under sat_count

    """
    if np.any(max_vals < sat_count):
        asort = np.argsort(max_vals)
        for i in range(1, len(asort)):
            if max_vals[asort[i]] >= sat_count:
                return asort[i - 1]
        return asort[i]
    else:
        return np.argmin(max_vals)


# This is actually not used.
# I've implemented *find_each_order_max* for extracting info for combining FLAT
def extract_ap_ridge_counts(im, ap_interp):
    npix = im.shape[1]
    naps = ap_interp.shape[0]

    # float -> int
    ap_int = ap_interp.astype(int)

    spec = np.zeros(ap_int.shape, dtype=float)
    sat_mask = np.zeros(ap_int.shape, dtype=bool)
    for j in range(npix):
        for i in range(naps):
            yc = ap_int[i, j]
            data = im[yc, j]
            spec[i, j] = data.data
    return spec, sat_mask


def find_each_order_max(im, ind_order, p=95):
    """ find the max for each order

    Actually the max value is affected by cosmic rays.
    So, finding the 99/95/90 percentiles could be more useful.


    Parameters
    ----------
    im:
        FLAT image
    ind_order:
        ind of apertures
    p: float
        percentile

    Returns
    -------
    max_list: ndarray


    """
    uorder = np.unique(ind_order)
    uorder_valid = uorder[uorder >= 0]

    max_list = np.zeros_like(uorder_valid)
    for each_uorder in uorder_valid:
        max_list[each_uorder] = np.percentile(
            im[np.where(ind_order == each_uorder)], p)
        print("@TWODSPEC: finding maximum value for order [%s]" % each_uorder)

    return max_list


# ################################# #
#      scatterd light correction
# ################################# #

def substract_scattered_light(im_, ap_uorder_interp, ap_width=10,
                              method='gaussian', method_kwargs=None,
                              shrink=.85):  # 10 should be max

    im = ccdproc.CCDData(copy(im_))
    assert method in {'gaussian', 'median'}

    ind_inter_order = find_inter_order(im, ap_uorder_interp, ap_width=ap_width)

    # fill edge with inter-order values
    for i_col in range(im.shape[1]):
        ind = np.where(ind_inter_order[:, i_col])[0]
        # print(ind)
        im.data[0:ind[0], i_col] = np.median(im[ind[0]:ind[0]+3, i_col])
        im.data[ind[-1]:, i_col] = np.median(im[ind[-1]-3:ind[-1], i_col])
    # plt.figure()
    # plt.imshow(im, interpolation='nearest')

    # initiate scattered light
    im_scattered_light = np.zeros_like(im)

    # interpolation
    x = np.arange(im.shape[0])
    print("@TWODSPEC: interpolating scattered light ...")
    for i_col in range(im.shape[1]):
        w = ind_inter_order[:, i_col]
        y = im[:, i_col]
        I = interp1d(x[w], y[w], 'linear', bounds_error=False, fill_value=0.)
        im_scattered_light[:, i_col] = I(x)

    # smooth
    print("@TWODSPEC: smoothing scattered light ...")
    # print(type(im_scattered_light))
    # print(im_scattered_light)
    # print(method_kwargs)
    if method is 'gaussian':
        im_scattered_light_smooth = gaussian(im_scattered_light * shrink, **method_kwargs)
    elif method is 'median':
        im_scattered_light_smooth = medfilt2d(im_scattered_light * shrink, **method_kwargs)

    return im - im_scattered_light_smooth, im_scattered_light_smooth


def find_inter_order(im, ap_uorder_interp, ap_width=10):
    ind_inter_order = np.ones_like(im, dtype=bool)

    # generate coordinates
    x, y = np.arange(im.shape[1]), np.arange(im.shape[0])
    mx, my = np.meshgrid(x, y)

    # for each apertures
    for i_order in range(ap_uorder_interp.shape[0] - 1):
        this_order_ridge = ap_uorder_interp[i_order, :].reshape(1, -1)
        ind_this_order = np.logical_and(my > this_order_ridge - ap_width,
                                        my < this_order_ridge + ap_width)
        ind_inter_order = np.where(ind_this_order, False, ind_inter_order)
        print ("@TWODSPEC: *find_inter_order* marking pixels of order [%s]" %
               i_order)
    # # upper edge
    # i_order = 0
    # upper_dist_typical = np.median(np.diff(ap_uorder_interp, axis=0)[0])
    # upper_dist_use = np.int(np.max([upper_dist_typical - 2 * ap_width, 3]))
    # this_order_ridge = ap_uorder_interp[i_order, :].reshape(1, -1)
    # ind_this_order = my < (this_order_ridge - ap_width - upper_dist_use)
    # ind_inter_order = np.where(ind_this_order, False, ind_inter_order)
    #
    # # lower edge
    # i_order = -1
    # lower_dist_typical = np.median(np.diff(ap_uorder_interp, axis=0)[0])
    # lower_dist_use = np.int(np.max([lower_dist_typical - 2 * ap_width, 3]))
    # this_order_ridge = ap_uorder_interp[i_order, :].reshape(1, -1)
    # ind_this_order = my > (this_order_ridge + ap_width + lower_dist_use)
    # ind_inter_order = np.where(ind_this_order, False, ind_inter_order)

    # bounds
    # ind_inter_order[0, :] = True
    # ind_inter_order[-1, :] = True
    # plt.figure()
    # plt.imshow(ind_inter_order, interpolation='nearest')
    # print(ind_inter_order)
    return ind_inter_order


# ################################# #
#      sensitivity variation correction
# ################################# #

def apflatten(flat, ap_uorder_interp, ap_width=(-8, 8), pct=50, **normalization):
    """ IRAF apflatten replication """

    # symmetric case
    if not isinstance(ap_width, tuple):
        ap_width = (-ap_width, ap_width)
    ap_npix = ap_width[1] - ap_width[0] + 1

    # fix apertures
    n_order, disp_npix = ap_uorder_interp.shape

    # get slice along dispersion axis
    order, ofst, xcoord, ycoord = get_aperture_index(ap_uorder_interp,
                                                     ap_width=ap_width)
    # slice 2D FLAT
    flat_slc = flat[ycoord, xcoord]

    # get 1d FLAT by sextract method (simple sum)
    flat1d_simple = sextract_all_aperture(
        flat, ap_uorder_interp, ap_width=ap_width)
    # fit 1d FLAT
    flat1d_cont = normalize_spectra_block(
        xcoord[0], flat1d_simple, norm_range=xcoord[0, [0, -1]], **normalization)[1]
    # tile shape
    flat_slc_cont = np.tile(
        flat1d_cont, (1, ap_npix)).reshape(ap_npix * n_order, -1)
    # caculate profile -> for FLAT
    flat1d_profile = np.percentile(flat_slc.data / flat_slc_cont, pct, axis=1)
    # reconstruct sliced 2d FLAT model
    flat_slc_model = flat1d_profile.reshape(-1, 1) * flat_slc_cont
    # normalize 2d FLAT to its model
    flat_slc_norm = flat_slc / flat_slc_model

    # caculate profile -> for STAR
    flat_slc_sum = np.tile(flat1d_simple, (1, ap_npix)).reshape(-1, disp_npix)
    star_slc_profile = flat_slc.data / flat_slc_sum

    # put back data
    flat2d_norm = np.ones_like(flat)
    flat2d_model = np.ones_like(flat)
    flat2d_norm[ycoord, xcoord] = flat_slc_norm
    flat2d_model[ycoord, xcoord] = flat_slc_model

    profile_slc = np.tile(flat1d_profile.reshape(-1, 1), (1, disp_npix))
    profile2d = np.zeros_like(flat)
    profile2d[ycoord, xcoord] = profile_slc

    return dict(
        flat=flat,
        flat_slc=flat_slc,
        flat1d_simple=flat1d_simple,
        flat1d_cont=flat1d_cont,
        flat_slc_cont=flat_slc_cont,
        flat1d_profile=flat1d_profile,
        flat_slc_model=flat_slc_model,
        flat_slc_norm=flat_slc_norm,
        flat2d_norm=flat2d_norm,
        flat2d_model=flat2d_model,
        profile_slc=profile_slc,
        profile2d=profile2d,
        star_slc_profile=star_slc_profile,
    )


def select_a_specific_aperture(order, ofst, x_coord, y_coord, specific_order=0):
    """ select a specific aperture """

    # determine sub-index
    ind = np.where(order == specific_order)[0]

    return order[ind], ofst[ind], x_coord[ind], y_coord[ind]


def sextract_a_specific_aperture(img, ap_uorder_interp, ap_width=(-8, 8), specific_order=0, func=np.sum):
    """ simple extraction for a specific order """

    # symmetric case
    if not isinstance(ap_width, tuple):
        ap_width = (-ap_width, ap_width)

    # fix apertures
    ap_uorder_interp = ap_uorder_interp.astype(int)

    # get slice along dispersion axis
    order, ofst, xcoord, ycoord = get_aperture_index(
        ap_uorder_interp, ap_width=ap_width)
    order, ofst, xcoord, ycoord = select_a_specific_aperture(
        order, ofst, xcoord, ycoord, specific_order=specific_order)

    # extract using a simple sum
    spec1d = func(img[ycoord, xcoord], axis=0)

    return spec1d


def sextract_all_aperture(img, ap_uorder_interp, ap_width=(-8, 8), func=np.sum):
    """ simple extraction for all orders """

    # symmetric case
    if not isinstance(ap_width, tuple):
        ap_width = (-ap_width, ap_width)

    # fix apertures
    ap_uorder_interp = ap_uorder_interp.astype(int)

    # get slice along dispersion axis
    order, ofst, xcoord, ycoord = get_aperture_index(
        ap_uorder_interp, ap_width=ap_width)

    # initiate spec1d
    spec1d = np.zeros(ap_uorder_interp.shape)

    # simple extraction
    for specific_order in range(ap_uorder_interp.shape[0]):
        try:
            # get index of this aperture
            order_, ofst_, xcoord_, ycoord_ = select_a_specific_aperture(
                order, ofst, xcoord, ycoord, specific_order=specific_order)
            # extract using a simple sum
            spec1d[specific_order] = func(img[ycoord_, xcoord_], axis=0)
        except:
            print("@SONG: problems occur, specific_order: ", specific_order)
            print(ycoord_, xcoord_)
            spec1d[specific_order] = np.nan

    return spec1d


def get_aperture_index(ap_uorder_interp, ap_width=(-8, 8)):
    """ return X & Y index for all apertures """

    # fix pixel index
    ap_uorder_interp = ap_uorder_interp.astype(int)

    # number of apertures
    n_order = ap_uorder_interp.shape[0]
    npix_apwidth = ap_width[1] - ap_width[0] + 1
    npix_aplength = ap_uorder_interp.shape[1]

    # initiate X & Y
    x_coord = np.arange(npix_aplength, dtype=int).reshape(1, -1).repeat(npix_apwidth*n_order, axis=0)
    y_coord = np.zeros((npix_apwidth*n_order, npix_aplength), dtype=int)

    # determine Y
    for i_order in np.arange(ap_uorder_interp.shape[0], dtype=int):
        row_start = i_order*npix_apwidth
        ofst = np.arange(ap_width[0], ap_width[1]+1, dtype=int).reshape(-1, 1)
        y_coord[row_start:row_start+npix_apwidth] = ap_uorder_interp[i_order] + ofst

    # determine order & ofst
    order = np.arange(n_order, dtype=int).repeat(npix_apwidth)
    ofst = np.arange(ap_width[0], ap_width[1] + 1, dtype=int).reshape(1, -1).repeat(n_order, axis=0).reshape((-1,))

    return order, ofst, x_coord, y_coord


# ################################# #
#      apbackground
# ################################# #

def apbackground(img, ap_uorder_interp, offsetlim=(-5, 5), q=50,
                 npix_inter=3, kernel_size=(11, 11), polydeg=5):
    """ determine background/scattered light using inter-aperture pixels """

    img = np.array(img)

    # determine inter-aperture pixels
    ap_lo = np.sum(np.array([3, -3, 1]).reshape(-1, 1) * ap_uorder_interp[0:3], axis=0)
    ap_hi = np.sum(np.array([1, -3, 3]).reshape(-1, 1) * ap_uorder_interp[-3:], axis=0)

    # risky to extend inter-order region!
    # ap_extended = np.vstack((ap_lo, ap_uorder_interp, ap_hi))
    # ap_inter = (ap_extended[1:] + ap_extended[:-1]) * .5
    ap_extended = np.vstack((ap_lo, ap_uorder_interp))
    ap_inter = (ap_extended[1:] + ap_extended[:-1]) * .5

    # determine the valley of interp-aperture pixels
    try:
        ap_width, sys_offset = apwidth(img, ap_inter, offsetlim=(-3, 3),
                                       ap_npix=npix_inter, method='min')
    except:
        ap_width = (-2, +2)

    # get slice along dispersion axis
    order, ofst, xcoord, ycoord = get_aperture_index(ap_inter, ap_width=ap_width)

    # extract median interp-aperture pixels
    def f(*args, **kwargs):
        return np.percentile(*args, **kwargs, q=q)

    spec1d_inter = sextract_all_aperture(img, ap_inter, ap_width=ap_width, func=f)
    spec_slc_tiled = np.tile(spec1d_inter, (1, ap_width[1] - ap_width[0] + 1)
                             ).reshape(-1, ap_inter.shape[1])

    # interpolate median background
    sl = np.zeros_like(img) * np.nan
    for i_col in range(sl.shape[1]):
        # xcoord_ = xcoord[:, i_col]
        ycoord_ = ycoord[:, i_col]
        sl_ = spec_slc_tiled[:, i_col]
        sl_interp = interp1d(ycoord_[np.isfinite(sl_)], sl_[np.isfinite(sl_)], bounds_error=False)(np.arange(sl.shape[0]))

        # fill the ends
        ind_end = np.where(np.isfinite(sl_interp))[0][[0, -1]]
        sl_interp[:ind_end[0]] = sl_interp[ind_end[0]]
        sl_interp[ind_end[-1] + 1:] = sl_interp[ind_end[-1]]
        sl[:, i_col] = sl_interp

    # median filter
    sls = medfilt2d(sl, kernel_size=kernel_size)

    # 1D polyfit smoothing
    print("@SONG: polyfit filtering ...")
    bg_polymoothed0 = polyfitfilt(sl, deg=polydeg, axis=0)
    print("@SONG: polyfit filtering2 ...")
    bg_polymoothed01 = polyfitfilt(bg_polymoothed0, deg=polydeg, axis=1)

    res = dict(
        bg=sl,
        bg_smoothed=sls,
        bg_poly=bg_polymoothed01
    )

    return res


def apbackground_poly(im, ap_uorder_interp, q=50, npix_inter=2, polydeg=5):
    mx, my = np.meshgrid(np.arange(im.shape[1]), np.arange(im.shape[0]))
    bg = np.zeros_like(im, dtype=float) * np.nan
    ap_inter = (ap_uorder_interp[1:] + ap_uorder_interp[:-1])/2.
    xcoord = np.arange(im.shape[1], dtype=int)
    ycoord = np.arange(im.shape[0], dtype=int)

    for ix in range(im.shape[1]):
        for i_inter in range(ap_inter.shape[0]):
            ycoord_ = np.arange(ap_inter[i_inter, ix] - npix_inter,
                                ap_inter[i_inter, ix] - npix_inter + 1,
                                1, dtype=int)
            bg[ycoord_, ix] = np.percentile(im[ycoord_, ix], q=q)

    for ix in range(im.shape[1]):
        ind_good = np.isfinite(bg[:, ix])
        I = interp1d(ycoord[ind_good], bg[ind_good, ix], bounds_error=False)
        bg[:, ix] = I(ycoord)
    # 1D polyfit smoothing
    bg_polymoothed0 = polyfitfilt(bg, deg=polydeg, axis=0)
    bg_polymoothed01 = polyfitfilt(bg_polymoothed0, deg=polydeg, axis=1)

    return bg_polymoothed01


def polyfitfilt(im, deg=5, axis=0):
    im = np.array(im)
    im_smoothed = np.zeros_like(im, dtype=float)
    if axis == 0:
        x = np.arange(im.shape[0])
        for i_col in range(im.shape[1]):
            y = im[:, i_col]
            ind_good = np.isfinite(y)
            p = np.polynomial.chebyshev.chebfit(x[ind_good], y[ind_good], deg=deg)
            im_smoothed[:, i_col] = np.polynomial.chebyshev.chebval(x, p)
    elif axis == 1:
        x = np.arange(im.shape[1])
        for i_row in range(im.shape[0]):
            y = im[i_row, :]
            ind_good = np.isfinite(y)
            p = np.polynomial.chebyshev.chebfit(x[ind_good], y[ind_good], deg=deg)
            im_smoothed[i_row, :] = np.polynomial.chebyshev.chebval(x, p)
    else:
        raise (ValueError("@SONG: Invalid axis value: ", axis))

    return im_smoothed


# ################################# #
#      apwidth
# ################################# #

def apwidth(img, ap_uorder_interp, offsetlim=(-8, 8), ap_npix=10, method='max'):
    """ automatically find ap_width for a given ap_npix """

    img = np.array(img)

    ap_width_max = offsetlim[1] - offsetlim[0] + 1
    assert 1 <= ap_npix <= ap_width_max

    # determine SNR of each offset
    ofst, medsnr_lnsum = apoffset_snr(img, ap_uorder_interp, offsetlim=offsetlim)

    if method is "max":
        istart = np.argmax(medsnr_lnsum)
        sys_offset = ofst[istart]

        i_ap_width_lo, i_ap_width_hi = istart, istart
        i_ap_width_min, i_ap_width_max = 0, len(ofst) - 1

        while (i_ap_width_hi - i_ap_width_lo + 1) < ap_npix:
            if i_ap_width_lo - 1 < i_ap_width_min:
                # towards high end
                i_ap_width_hi += 1
            elif i_ap_width_hi + 1 > i_ap_width_max:
                # towards low end
                i_ap_width_lo -= 1

                # then index is inside
            elif medsnr_lnsum[i_ap_width_hi + 1] >= medsnr_lnsum[
                        i_ap_width_lo - 1]:
                # towards high end
                i_ap_width_hi += 1
            elif medsnr_lnsum[i_ap_width_hi + 1] < medsnr_lnsum[
                        i_ap_width_lo - 1]:
                # towards low end
                i_ap_width_lo -= 1
            else:
                raise ValueError(
                    "@SONG: value error when automatically finding ap_width\n"
                    "{0}".format(medsnr_lnsum.__repr__()))
    elif method is "min":
        istart = np.argmin(medsnr_lnsum)
        sys_offset = ofst[istart]

        i_ap_width_lo, i_ap_width_hi = istart, istart
        i_ap_width_min, i_ap_width_max = 0, len(ofst) - 1

        while (i_ap_width_hi - i_ap_width_lo + 1) < ap_npix:
            if i_ap_width_lo - 1 < i_ap_width_min:
                # towards high end
                i_ap_width_hi += 1
            elif i_ap_width_hi + 1 > i_ap_width_max:
                # towards low end
                i_ap_width_lo -= 1

                # then index is inside
            elif medsnr_lnsum[i_ap_width_hi + 1] <= medsnr_lnsum[
                        i_ap_width_lo - 1]:
                # towards high end
                i_ap_width_hi += 1
            elif medsnr_lnsum[i_ap_width_hi + 1] > medsnr_lnsum[
                        i_ap_width_lo - 1]:
                # towards low end
                i_ap_width_lo -= 1
            else:
                raise ValueError(
                    "@SONG: value error when automatically finding ap_width\n"
                    "{0}".format(medsnr_lnsum.__repr__()))
    else:
        raise ValueError("@SONG: invalid method! method={0}".format(method))

    ap_width = (ofst[i_ap_width_lo], ofst[i_ap_width_hi])

    return ap_width, sys_offset


# ################################# #
#      apoffset
# ################################# #

def apoffset_snr(img, ap_uorder_interp, offsetlim=(-8, 8)):
    """ determine the signal-to-noise ratio for offset """
    img = np.array(img)
    n_offset = offsetlim[1] - offsetlim[0] + 1

    # Poisson error
    var = np.abs(img)
    snr = img/np.sqrt(var)

    # get index
    order, ofst, xcoord, ycoord = get_aperture_index(
        ap_uorder_interp, ap_width=offsetlim)

    # median signal-to-noise ratio for each offset
    medsnr = np.nanmedian(snr[ycoord, xcoord], axis=1)

    medsnr_lnsum = np.zeros((n_offset,), dtype=float)
    for i in np.arange(n_offset, dtype=int):
        medsnr_lnsum[i] = np.nansum(np.log(medsnr[i::n_offset]))

    ofst = np.arange(offsetlim[0], offsetlim[1] + 1, dtype=int)

    return ofst, medsnr_lnsum


# ################################# #
#      apvariance
# ################################# #

def apvariance():
    pass


# ################################# #
#      apsaturation
# ################################# #

def apsaturation(img, adulim=65536., gain=4.0):
    # return the saturation mask of img
    return img > (adulim*gain)


# ################
def get_dispersion_slice_index(ap_uorder_interp, i_order, offset=0):
    indx = np.arange(ap_uorder_interp.shape[1], dtype=int)
    indy = np.array(ap_uorder_interp[i_order]+offset, dtype=int)
    return indy, indx


def get_dispersion_slice(im, ap_uorder_interp, i_order, offset=0):
    indy, indx = get_dispersion_slice_index(
        ap_uorder_interp, i_order, offset=offset)
    return np.array(im)[indy, indx]


def put_dispersion_slice(data, im, ap_uorder_interp, i_order, offset=0):
    # get index
    indy, indx = get_dispersion_slice_index(
        ap_uorder_interp, i_order, offset=offset)

    # put data
    im_ = np.array(copy(im))
    im_[indy, indx] = data
    return im_




import numpy as np
from scipy.optimize import least_squares


def cost1d(p, x, y, pw=2):
    res = np.polynomial.chebyshev.chebval(x, p) - y
    return np.abs(res) ** (pw / 2)


def cost2d(p, x, y, z, deg=(1, 1), pw=2):
    res = np.polynomial.chebyshev.chebval2d(x, y, p.reshape(deg[0]+1, deg[1]+1)) - z
    return np.abs(res) ** (pw / 2)


class Poly1DFitter:
    x_mean = 0
    x_scale = 0
    y_mean = 0
    y_scale = 0

    x_sc = 0
    y_sc = 0

    opt = None
    p = None
    deg = 0
    rms = 0

    def __init__(self, x, y, deg=1, pw=1, robust=True):
        # determine data scale
        if robust:
            self.x_scale = np.ptp(np.percentile(x, q=[84, 16])) / 2
            self.x_mean = np.percentile(x, q=50)
            self.y_scale = np.ptp(np.percentile(y, q=[84, 16])) / 2
            self.y_mean = np.percentile(y, q=50)
        else:
            self.x_scale = np.std(x)
            self.x_mean = np.mean(x)
            self.y_scale = np.std(y)
            self.y_mean = np.mean(y)

        # normalize data
        self.x_sc = (x - self.x_mean) / self.x_scale
        self.y_sc = (y - self.y_mean) / self.y_scale

        # fit data
        p0 = np.zeros((deg + 1,), dtype=float)
        self.opt = least_squares(cost1d, p0, method="lm", args=(self.x_sc, self.y_sc, pw))
        self.p = self.opt.x

        self.pw = pw
        self.robust = robust
        self.deg = deg

    def predict(self, x):
        x_sc = (x - self.x_mean) / self.x_scale
        y_sc = np.polynomial.chebyshev.chebval(x_sc, self.p)
        return y_sc * self.y_scale + self.y_mean


class Poly2DFitter:
    x_mean = 0
    x_scale = 0
    y_mean = 0
    y_scale = 0
    z_mean = 0
    z_scale = 0

    x_sc = 0
    y_sc = 0
    z_sc = 0

    opt = None
    p = None
    deg = 0
    rms = 0

    def __init__(self, x, y, z, deg=(1, 2), pw=1, robust=True):
        # determine data scale
        if robust:
            self.x_scale = np.ptp(np.percentile(x, q=[84, 16])) / 2
            self.x_mean = np.percentile(x, q=50)
            self.y_scale = np.ptp(np.percentile(y, q=[84, 16])) / 2
            self.y_mean = np.percentile(y, q=50)
            self.z_scale = np.ptp(np.percentile(z, q=[84, 16])) / 2
            self.z_mean = np.percentile(z, q=50)
        else:
            self.x_scale = np.std(x)
            self.x_mean = np.mean(x)
            self.y_scale = np.std(y)
            self.y_mean = np.mean(y)
            self.z_scale = np.std(z)
            self.z_mean = np.mean(z)

        # normalize data
        self.x_sc = (x - self.x_mean) / self.x_scale
        self.y_sc = (y - self.y_mean) / self.y_scale
        self.z_sc = (z - self.z_mean) / self.z_scale

        # fit data
        p0 = np.zeros((deg[0] + 1, deg[1] + 1), dtype=float).flatten()
        self.opt = least_squares(cost2d, p0, method="lm", args=(self.x_sc, self.y_sc, self.z_sc, deg, pw))
        self.p = self.opt.x

        self.pw = pw
        self.robust = robust
        self.deg = deg

    def predict(self, x, y):
        x_sc = (x - self.x_mean) / self.x_scale
        y_sc = (y - self.y_mean) / self.y_scale
        z_sc = np.polynomial.chebyshev.chebval2d(
            x_sc, y_sc, self.p.reshape(self.deg[0]+1, self.deg[1]+1))
        return z_sc * self.z_scale + self.z_mean
